(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import Arith.
Require Import Base.
Require Import List.
Require Import Setoid.

Require Import Maps.
Require Import Machine.
Require Import ModelSpec.
Require Import Worlds.

Module LMWorldCarrier <: WORLD_CARRIER_WITH_HOLE.
  
  Record lm_world (A:Type) : Type := World {
   world_s: store;
   world_v: value;
   world_hole : A
  }.

  Definition world_functor := lm_world.

  Definition upd_s A (s: store) (w: lm_world A) := match w with World _ v h => World A s v h end.
  Definition upd_v A (v: value) (w: lm_world A) := match w with World s _ h => World A s v h end.
  Definition upd_h A B (h: B)   (w: lm_world A) := match w with World s v _ => World B s v h end.

  Implicit Arguments world_s.
  Implicit Arguments world_v.
  Implicit Arguments world_hole.
  Implicit Arguments upd_s.
  Implicit Arguments upd_v.

  Definition sample_world A (x:A) :=
     World A (empty _) (value_label (L 0)) x.

  Inductive field' : Set := Sel_S | Sel_V.

  Definition field := field'.

  Definition field_enum (f:field) : nat :=
    match f with
    | Sel_S => 0
    | Sel_V => 1
    end.

  Lemma field_enum_injective : forall x y,
     field_enum x = field_enum y -> x = y.

    intros x y; destruct x; destruct y; simpl; intuition; discriminate.
   Qed.

  Definition field_eq_dec : forall x y:field, {x=y}+{x<>y}.
    decide equality.
   Defined.

  Definition field_type (f:field) : Type :=
    match f with
    | Sel_S => store
    | Sel_V => value
    end.

  Definition w_upd (A:Type) (f:field) : field_type f -> lm_world A -> lm_world A :=
    match f as f' return field_type f' -> lm_world A -> lm_world A with
    | Sel_S => @upd_s A
    | Sel_V => @upd_v A
    end.

  Definition w_proj (A:Type) (f:field) : lm_world A -> field_type f :=
    match f as f' return lm_world A -> field_type f' with
    | Sel_S => @world_s A
    | Sel_V => @world_v A
    end.

  Lemma proj_upd_same : forall A x a w,
     w_proj A x (w_upd A x a w) = a.

     intros A x a w; destruct x; destruct w; simpl; auto.
    Qed.

  Lemma proj_upd_diff : forall A x y a w,
     x <> y ->
     w_proj A x (w_upd A y a w) = w_proj A x w.

     intros A x y a w; destruct x; destruct y; destruct w; simpl; auto;
       intros; contradiction H; auto.
    Qed.

  Lemma upd_collapse : forall A x a b w,
     w_upd A x a (w_upd A x b w) =
     w_upd A x a w.

     intros A x a b w; destruct x; destruct w; simpl; auto.
    Qed.

  Lemma upd_commute : forall A x y a b w,
     x <> y ->
     w_upd A x a (w_upd A y b w) =
     w_upd A y b (w_upd A x a w).

     intros A x y a b w; destruct x; destruct y; destruct w; simpl; auto;
       intros; contradiction H; auto.
    Qed.

  Lemma upd_proj_same : forall A x w,
     w_upd A x (w_proj A x w) w = w.

     intros A x w; destruct x; destruct w; simpl; auto.
    Qed.

  Lemma upd_injective : forall A x a a' w w',
     w_upd A x a w = w_upd A x a' w' -> a = a'.

     intros; destruct x; destruct w; destruct w'; simpl in *; congruence.
    Qed.

  Lemma upd_eq : forall A x a b w w',
     w_upd A x a w = w_upd A x a w' ->
     w_upd A x b w = w_upd A x b w'.

     intros; destruct x; destruct w; destruct w'; simpl in *;
       injection H; intros; subst; auto.
   Qed.

  Definition get_hole := world_hole.
  Definition set_hole := upd_h.

  Lemma get_set_hole : forall A B x w,
     get_hole B (set_hole A B x w) = x.

     intros; destruct w; simpl; auto.
    Qed.

  Lemma set_get_hole_same : forall A w,
     set_hole A A (get_hole A w) w = w.

     intros; destruct w; simpl; auto.
    Qed.

  Lemma set_hole_collapse : forall A B C x y w,
     set_hole B C y (set_hole A B x w) =
     set_hole A C y w.

     intros; destruct w; auto.
    Qed.

  Lemma set_hole_injective : forall A B x x' w w',
     set_hole A B x w = set_hole A B x' w' -> x = x'.

     intros; destruct w; destruct w'; injection H; auto.
    Qed.

  Lemma set_hole_eq : forall A B x y w w',
     set_hole A B x w = set_hole A B x w' ->
     set_hole A B y w = set_hole A B y w'.

     intros; destruct w; destruct w'; simpl in *;
       injection H; intros; subst; auto.
    Qed.

  Lemma set_hole_commute : forall A B x a b w,
     w_upd B x a (set_hole A B b w) =
     set_hole A B b (w_upd A x a w).

     intros; destruct x; destruct w; simpl in *; auto.
    Qed.

  Lemma get_hole_upd_collapse : forall A x a w,
    get_hole A (w_upd A x a w) = get_hole A w.

    intros.
    rewrite <- (set_get_hole_same _ w).
    rewrite set_hole_commute.
    rewrite get_set_hole.
    rewrite (set_get_hole_same _ w).
    auto.
   Qed.

End LMWorldCarrier.

Module LMSimpleWorlds := SimpleWorlds LMWorldCarrier.

Module LMStep <: WORLD_STEP with Module W:=LMSimpleWorlds.
  Module W:=LMSimpleWorlds.

  Import LMWorldCarrier.
  Import LMSimpleWorlds.WC.
  Import LMSimpleWorlds.WT.

  Definition prog_desc : Type := program.
  Definition prog_cont : Type := instr.

  Definition step (p:program) (i1:instr) (w1:world) (i2:instr) (w2:world) :=
     Machine.step p (world_s w1) i1 (world_s w2) i2
     /\ world_v w1 = world_v w2 /\ world_hole w1 = world_hole w2.

  Definition halted (p:program) (i:instr) (w:world) := i = instr_halt.

  Lemma step_deterministic : forall p k k1 k2 w w1 w2,
     step p k w k1 w1 -> step p k w k2 w2 -> k1  = k2 /\ w1 = w2.

     unfold step; intros.
     destruct w; destruct w1; destruct w2; simpl in *.
     decompose [and] H; clear H.
     decompose [and] H0; clear H0.
     generalize (step_deterministic _ _ _ _ _ _ _ H1 H).
     intuition; congruence.
    Qed.

  Lemma step_preserves_ttl : forall p k k' w w',
     step p k w k' w' -> world_ttl w = world_ttl w'.

     unfold step, world_ttl; intuition.
    Qed.

  Lemma halted_no_step : forall p k k' w w',
     halted p k w -> step p k w k' w' -> False.

     unfold step, halted; intuition.
     rewrite H in H1.
     inversion H1.
    Qed.

  Lemma halt_tick : forall p k w w',
     tick_world w = Some w' ->
     (halted p k w <-> halted p k w').

     unfold halted, tick_world; intros.
     intuition.
    Qed.

  Lemma step_tick_commute : forall p k k' w w',
     (exists w1, step p k w k' w1 /\ tick_world w1 = Some w') <->
     (exists w2, tick_world w = Some w2 /\ step p k w2 k' w').

    split; intros; destruct H;
      unfold step, tick_world, world_ttl, get_hole in *; intuition.

    case_eq (world_hole x); intros; rewrite H2 in H1; try discriminate.
    case_eq (world_hole w); intros.
    absurd (0=S n); auto; congruence.
    replace n0 with n in *; [ | congruence ].
    exists (upd_h _ _ n w); intuition;
      destruct w; destruct w'; destruct x; simpl in *;
        congruence.
    case_eq (world_hole w); intros; rewrite H2 in H0; try discriminate.
    exists (upd_h _ _ (S n) w'); intuition;
      destruct w; destruct w'; destruct x; simpl in *;
        congruence.
   Qed.
End LMStep.

Module Type LM_MODAL_MODEL := MODAL_MODEL
 with Module W := LMSimpleWorlds
 with Module WS := LMStep.
