(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy *)
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 * )

Require Export Arith.
Require Export Bool.
Require Export RecType.
Require Export ProgramSafety.
Require Export InstrRules.
Require Export Listmachine.

Require Import Maps.

Require Import CompleteInduction.

Open Local Scope ty_scope.

Inductive list_ty: Set :=
  | lty_top : list_ty
  | lty_bot : list_ty
  | lty_nil: list_ty
  | lty_list: list_ty -> list_ty
  | lty_listcons: list_ty -> list_ty
  | lty_cont : map list_ty -> list_ty
  .

Definition contmap := map list_ty.
Definition program_typing := map contmap.

Fixpoint list_ty_depth (x:list_ty) : nat :=
  match x with
  | lty_top => 1
  | lty_bot => 1
  | lty_nil => 1
  | lty_list x' => 1 + (list_ty_depth x')
  | lty_listcons x' => 1 + (list_ty_depth x')
  | lty_cont m => 1 + 
     (map_fold_index _ _ (fun _ x => list_ty_depth x) plus 0 m)
  end.

Fixpoint interp_list_ty (p:program) (t:list_ty) { struct t } : ty :=
  match t with
  | lty_top => t_top
  | lty_bot => t_bot
  | lty_nil => t_nil
  | lty_list t => t_list (interp_list_ty p t)
  | lty_listcons t => t_listcons (interp_list_ty p t)
  | lty_cont m => t_cont p
     (map_fold_index _ _ (fun i x => t_slot i (interp_list_ty p x)) t_intersection t_top m)
  end.

Definition interp_cont (p:program) (m:contmap) := 
   map_fold_index _ _ (fun i x => t_slot i (interp_list_ty p x)) t_intersection t_top m.
  
Notation "[[  t | p ]" := (interp_list_ty p t) : ty_scope.




Lemma interp_ignore_nec : forall p n,
   (forall c x, contmap_depth c < n ->
     interp_cont p c x |-- ignoring (Sel_I :: Sel_V :: nil) (interp_cont p c x) /\
     interp_cont p c x |-- |>(interp_cont p c x)) /\
   (forall t, list_ty_depth t < n ->
     [[ t |p] |-- ignoring (Sel_I::Sel_S::nil) [[ t |p] /\
     [[ t |p] |-- |> [[ t |p]).

   intros p n; pattern n; apply CompleteInduction.
   clear n; intros n Hind.
   destruct n; [ split; intros; inversion H | ].
   destruct (Hind n); auto; clear Hind.
   intuition.

   destruct c; simpl in *.
   apply top_ignores.
   destruct o.
   apply intersection_ignores.
   change (ignores ((Sel_I::nil)++(Sel_V::nil)) (t_slot x [[ l |p])).
   apply ignores_app.
   apply slot_ignores.
   simpl; intuition; discriminate.
   destruct (H0 l).
   omega.
   apply ignores_incl with (Sel_I::Sel_S::nil).
   red; simpl; intuition.
   apply H2.
   apply slot_ignores_V.
   destruct (H c (x+1)%positive).
   omega.
   apply H2.
   destruct (H c (x+1)%positive).
   omega.
   apply H2.

   destruct c; simpl in *.
   apply_d (ignore_later (Sel_N::nil)).
   apply top_ignores.
   simpl; auto.
   destruct o.
   simpl_derives.
   destruct (H0 l).
   omega.
   apply_d slot_nec.
   destruct (H0 l).
   omega.
   auto.
   assumed.
   destruct (H c (x+1)%positive).
   omega.
   apply_d H3; assumed.
   destruct (H c (x+1)%positive).
   omega.
   apply_d H3; assumed.

   destruct t; simpl in *.
   apply top_ignores.
   apply bot_ignores.
   unfold t_nil.
   unfold t_just.
   apply projecting_ignores.
   simpl; intuition; discriminate.
   apply list_ignores.
   simpl; intuition; discriminate.
   simpl; intuition; discriminate.
   destruct (H0 t); auto.
   omega.
   unfold t_listcons.
   apply intersection_ignores.
   apply list_ignores.
   simpl; intuition; discriminate.
   simpl; intuition; discriminate.
   destruct (H0 t); auto.
   omega.
   unfold t_pair.
   apply comp_prop_ignores.
   simpl; intuition; discriminate.
   intros.
   destruct a; simpl.
   apply bot_ignores.
   apply intersection_ignores.
   apply with_subst_ignores.
   destruct (H0 t); auto.
   omega.
   apply with_subst_ignores.
   destruct (H0 t); auto.
   omega.
   apply later_ignores.
   simpl; intuition; discriminate.
   apply list_ignores.
   simpl; intuition; discriminate.
   simpl; intuition; discriminate.
   destruct (H0 t); auto.
   omega.
   unfold t_cont.
   apply exist_ignores; intros.
   apply intersection_ignores; intros.
   apply intersection_ignores; intros.
   apply exist_ignores; intros.
   unfold t_labelblock.
   apply const_ignores.
   unfold t_just.
   apply projecting_ignores.
   simpl; intuition; discriminate.
   apply later_ignores.
   simpl; intuition; discriminate.
   unfold t_sub.
   apply nec_ignores.
   simpl; intuition; discriminate.
   unfold t_every.
   apply ignoring_ignores_immediate.
   red; simpl; intuition.

   destruct t; simpl in *.   
   apply derives_axiom; apply modal_N; apply top_intro.
   elim_at 0 bot_elim.
   apply nil_nec.
   apply list_nec'.
   destruct (H0 t); auto.
   omega.
   unfold t_listcons.
   simpl_derives.
   apply_d list_nec'.
   assumed.
   destruct (H0 t); auto.
   omega.

   clear_at 0.
   apply pair_nec.
   destruct (H0 t); auto.
   omega.
   apply derives_modal.
   apply list_nec'.
   destruct (H0 t); auto.
   omega.
   apply cont_nec.
   destruct (H c 1%positive).
   omega.
   auto.
  Qed.


Definition cm_get (m:contmap) (x:positive) := 
  lookup_cm (nat_of_P x - 1) m.

Definition cm_set (m:contmap) (x:positive) (t:list_ty) :=
  setvalue_cm m (nat_of_P x - 1) t.

Definition cm_unset (m:contmap) (x:positive) :=
  unset_cm m (nat_of_P x - 1).

Lemma get_set_cm : forall x m t,
  cm_get (cm_set m x t) x = Some t.

  unfold cm_get, cm_set; intros; apply get_set_cm_nat.
 Qed.

Lemma nat_of_P_injective : forall x v, 
  nat_of_P v = nat_of_P x ->
  x = v.

  intros x v.
  case_eq ((x ?= v)%positive Eq); intros.
  apply Pcompare_Eq_eq; auto.
  absurd (nat_of_P x < nat_of_P v).
  omega.
  apply nat_of_P_lt_Lt_compare_morphism; auto.
  absurd (nat_of_P x > nat_of_P v).
  omega.
  apply nat_of_P_gt_Gt_compare_morphism; auto.
 Qed.  


Lemma lookup_cm_interp : forall p e v n t x,
   nat_of_P x <= nat_of_P v ->
   n = nat_of_P v - nat_of_P x ->
   lookup_cm n e = Some t ->
   |-- interp_cont p e x >=> t_slot v [[ t |p].

   induction e; simpl in *; intros.
   elimtype False.
   induction n; simpl in *; discriminate.
   
   destruct n; simpl in *.
   destruct o; try discriminate.
   unfold t_sub; simpl.
   repeat apply modal_N.
   simpl_derives.
   replace l with t; [ | congruence ].
   replace v with x; [ assumed | ].
   apply nat_of_P_injective.
   omega.   
   
   destruct o; simpl in *.
   cut_with (interp_cont p e (x+1) >=> 
     t_slot v [[ t |p]).
   unfold t_sub; do 2 coalesce_modal.
   elim_at 2 implies_elim; assumed.
   apply IHe with n; auto;
     rewrite <- Pplus_one_succ_r;
     rewrite nat_of_P_succ_morphism;
     omega.
   cut_with (interp_cont p e (x+1) >=> 
     t_slot v [[ t |p]).
   unfold t_sub; do 2 coalesce_modal.
   elim_at 1 implies_elim; assumed.
   apply IHe with n; auto;
     rewrite <- Pplus_one_succ_r;
     rewrite nat_of_P_succ_morphism;
     omega.
  Qed.      

   
Lemma cm_get_interp : forall p e v t,
   cm_get e v = Some t ->
   |-- interp_cont p e 1 >=> t_slot v [[ t |p].

   unfold cm_get; intros.
   apply lookup_cm_interp with (nat_of_P v - nat_of_P 1).
   generalize (ZL4 v).
   intros [z Hz].
   rewrite Hz.
   unfold nat_of_P; simpl.
   auto with arith.
   auto.
   unfold nat_of_P in *; simpl; auto.
  Qed.

Lemma cm_set_unset : forall e v t,
    cm_set e v t = cm_set (cm_unset e v) v t.
  
   unfold cm_set, cm_unset.
   intros; apply set_unset.
  Qed.  

Lemma setvalue_cm_interp : forall p n e v t x,
  nat_of_P x <= nat_of_P v ->
  n = nat_of_P v - nat_of_P x ->
  [ interp_cont p e x :: t_slot v [[ t |p] :: nil ] |--
   interp_cont p (setvalue_cm e n t) x.

  induction n; destruct e; try destruct o; simpl; intros.
  replace x with v.
  simpl_derives; assumed.
  apply nat_of_P_injective.
  omega.
  replace x with v.
  simpl_derives; try assumed.
  apply nat_of_P_injective.
  omega.
  replace x with v.
  simpl_derives; try assumed.
  apply nat_of_P_injective.
  omega.
  refine (IHn cm_nil v t (x+1)%positive _ _).
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  simpl_derives; try assumed.
  clear_at 0.
  apply IHn.  
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  apply IHn.  
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
 Qed.
   
Lemma cm_set_interp : forall p e v t,
  [ interp_cont p e 1 :: t_slot v [[ t |p] :: nil ] |--
   interp_cont p (cm_set e v t) 1.

  unfold cm_set.
  intros.
  apply setvalue_cm_interp.
  generalize (ZL4 v).
  intros [h Hh].
  rewrite Hh.
  compute; omega.
  compute; auto.
 Qed.

Lemma unset_cm_interp : forall p n e x,
  [ interp_cont p e x :: nil ] |--
    interp_cont p (unset_cm e n) x.

  induction n; destruct e; try destruct o; simpl; intros.
  apply top_intro.
  simpl_derives; assumed.
  assumed.
  apply top_intro.
  simpl_derives; try assumed.
  clear_at 0.
  apply IHn.
  apply IHn.
 Qed.

Lemma cm_unset_interp : forall p e v,
   interp_cont p e 1 |-- interp_cont p (cm_unset e v) 1.

   unfold cm_unset.
   intros.
   apply unset_cm_interp.
  Qed.


Lemma positive_add_neq : forall (x v z:positive),
   (x = v + z -> v <> x)%positive.

   intros.
   rewrite H.
   cut ((z + v)%positive <> v).
   rewrite Pplus_comm.
   auto.
   apply Pplus_no_neutral.
  Qed.

Lemma interp_cont_gt : forall p e x v z l s,
  (x = v + z)%positive ->
  [ projecting Sel_S s :: interp_cont p e x :: nil ] |--
    with_subst (existT SelType Sel_S (s#v <- l) ::nil) (interp_cont p e x).

  induction e; simpl; intros.
  apply derives_axiom.
  apply modal_N.
  apply top_intro.
  destruct o; simpl.
  simpl_derives.
  clear_at 1.
  move_at 1.
  unfold t_slot.
  simpl_derives.
  elim_at 0 (comp_prop_elim Sel_S s).
  assumed.
  simpl_derives.
  rewrite with_subst_distrib_exist.
  apply exist_intro with x0.
  simpl_derives.
  apply_d with_subst_prop.
  apply derives_axiom; apply modal_N.
  apply const_intro.
  rewrite get_set_other; auto.
  apply positive_add_neq with z.
  auto.
  in_goal rewrite discard_with_subst.
2:  simpl.
2:  apply with_subst_ignores.
2:  destruct (interp_ignore_nec p (1+list_ty_depth l0)).
2:  destruct (H2 l0); auto.
2:  apply ignores_incl with (Sel_I::Sel_S::nil).
2:  red; simpl; intuition.
2:  auto.
  assumed.
  clear_at 0.
  move_at 1.
  apply IHe with (Psucc z)%positive.
  rewrite <- Pplus_one_succ_r.
  rewrite Pplus_succ_permute_r.
  rewrite H; auto.
  apply IHe with (Psucc z)%positive.
  rewrite <- Pplus_one_succ_r.
  rewrite Pplus_succ_permute_r.
  rewrite H; auto.
 Qed.


Lemma unset_cm_not_in_dom : forall p n e v x,
  nat_of_P x <= nat_of_P v ->
  n = nat_of_P v - nat_of_P x ->
  not_in_dom v (interp_cont p (unset_cm e n) x).

  unfold not_in_dom, with_var_upd.
  intros.
  apply (with_projecting Sel_S); intro s.
  apply comp_prop_intro with s.
  assumed.  
  revert e v x H H0 v0 s.
  induction n; destruct e; try destruct o; simpl; intros.
  apply derives_axiom; apply modal_N; apply top_intro.
  replace v with x.
  apply interp_cont_gt with 1%positive; auto.
  apply nat_of_P_injective.
  omega.  
  replace v with x.
  apply interp_cont_gt with 1%positive; auto.
  apply nat_of_P_injective.
  omega.  
  apply derives_axiom; apply modal_N; apply top_intro.
  simpl_derives.
  
  unfold t_slot.
  simpl_derives.
  elim_at 0 (comp_prop_elim Sel_S s).
  assumed.
  simpl_derives.
  rewrite with_subst_distrib_exist.
  apply exist_intro with x0.
  simpl_derives.
  apply_d with_subst_prop.
  apply derives_axiom; apply modal_N.
  apply const_intro.
  rewrite get_set_other; auto.
  red; intros.
  subst v; omega.
  in_goal rewrite discard_with_subst.
2:  simpl.
2:  apply with_subst_ignores.
2:  destruct (interp_ignore_nec p (1+list_ty_depth l)).
2:  destruct (H3 l); auto.
2:  apply ignores_incl with (Sel_I::Sel_S::nil).
2:  red; simpl; intuition.
2:  auto.
  assumed.
  clear_at 0.
  move_at 1.
  apply IHn.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  apply IHn.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
  rewrite <- Pplus_one_succ_r.
  rewrite nat_of_P_succ_morphism.
  omega.
 Qed.

Lemma cm_unset_not_in_dom : forall p e v,
  not_in_dom v (interp_cont p (cm_unset e v) 1).

  unfold cm_unset; intros.
  apply unset_cm_not_in_dom.
  unfold nat_of_P at 1; simpl.
  destruct (ZL4 v).
  rewrite H; omega.
  compute; auto.  
 Qed.

Definition check_branch (p:program) (pt:program_typing) (vt:contmap) (l:label) : Prop :=
  match pt#l with
  | None => False
  | Some vt' => |-- interp_cont p vt 1 >=> interp_cont p vt' 1
  end.

Definition interp_pt (p:program) (pt:program_typing) : map ty :=
  map_fmap _ _ pt (fun x => interp_cont p x 1).
  
Definition lm_pt (p:program) (pt:program_typing) :=
  pgm_typ p (interp_pt p pt).

Definition lm_ok (p:program) (pt:program_typing) :=
  pt_domain_ok p (interp_pt p pt).

Lemma lm_pt_every : forall p pt,
  lm_pt p pt |-- !(lm_pt p pt).

  intros; unfold lm_pt, t_every.
  simpl_derives.
  apply pgm_typ_everywhere.
 Qed.

Lemma lm_ok_nec_every : forall p pt,
  lm_ok p pt |-- []!(lm_ok p pt).

  intros; unfold lm_ok.
  apply pt_domain_ok_nec_every.
 Qed.

Lemma lm_ok_labelblock : forall p pt l c,
  pt # l = Some c ->
  lm_ok p pt |-- Ex_ i:instr, t_labelblock p l i.

  intros; unfold lm_ok; simpl_derives.
  unfold t_labelblock.
  unfold pt_domain_ok.
  elim_at 0 (forall_elim _ l).
  elim_at 0 cut_elim_impl.
  apply exist_intro with (interp_cont p c 1).
  apply const_intro.
  unfold interp_pt.
  change (interp_cont p c 1) with ((fun x => interp_cont p x 1) c).
  apply fmap_eqn; auto.
   
  simpl_derives.
  apply exist_intro with x.
  apply const_intro; auto.
 Qed.  

Lemma lm_pt_labelguard : forall p pt l c,
  pt#l = Some c ->
  [ lm_pt p pt :: lm_ok p pt :: nil ] |-- labelguard p l (interp_cont p c 1).

  intros.
  unfold lm_pt, lm_ok, pgm_typ, pt_domain_ok.
  elim_at 0 (forall_elim _ l).
  elim_at 1 (forall_elim _ l).  
  elim_at 0 cut_elim_impl.
  apply exist_intro with (interp_cont p c 1).
  apply const_intro.
  unfold interp_pt.
  change (interp_cont p c 1) with ((fun x => interp_cont p x 1) c).
  apply fmap_eqn; auto.
  simpl_derives.
  elim_at 1 (forall_elim _ x); intros.
  elim_at 0 cut_elim_impl.
  unfold t_labelblock.
  apply const_intro; auto.
  simpl_derives.
  cut ((interp_pt p pt)#l = Some (interp_cont p c 1)).
  intros.
  replace x0 with (interp_cont p c 1).
  assumed.
  congruence.
  unfold interp_pt.
  change (interp_cont p c 1) with ((fun x => interp_cont p x 1) c).
  apply fmap_eqn; auto.
 Qed.
