(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy *)
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 * )

Require Export TC_syntax.
Require Export TC_rules.

Require Import CompleteInduction.

Open Local Scope ty_scope.

Fixpoint emptymap (m:contmap) : bool :=
  match m with
  | cm_nil => true
  | cm_cons None m' => emptymap m'
  | _ => false
  end.

Lemma emptymap_top : forall p m x,
  emptymap m = true ->
  |-- t_top >=> interp_cont p m x.

  induction m; simpl; intros.
  apply top_sub.
  destruct o; try discriminate.
  apply IHm; auto.
 Qed.  

Section check_sub_defs.
  Variables
  (co_sub : list_ty -> list_ty -> bool )
  (contra_sub : list_ty -> list_ty -> bool)
  (co_map_sub : contmap -> contmap -> bool)
  (contra_map_sub : contmap -> contmap -> bool).

  Definition check_sub' 
  (t1 t2:list_ty) : bool :=
  match t1, t2 with
  | _, lty_top => true
  | lty_bot, _ => true
  | lty_nil, lty_nil => true
  | lty_nil, lty_list t1 => true
  | lty_list s1, lty_list t1 => co_sub s1 t1
  | lty_listcons s1, lty_list t1 => co_sub s1 t1
  | lty_listcons s1, lty_listcons t1 => co_sub s1 t1
  | lty_cont m1, lty_cont m2 => contra_map_sub m2 m1
  | _, _ =>  false
  end.

  Definition check_map_sub' (m1 m2:contmap) : bool :=
  match m1, m2 with
  | _ , cm_nil                   => true
  | cm_nil, _                    => emptymap m2
  | cm_cons _ m, cm_cons None m' => co_map_sub m m'
  | cm_cons (Some t1) m, 
    cm_cons (Some t2) m'         => andb (co_sub t1 t2) (co_map_sub m m')
  | _, _ => false
  end.
End check_sub_defs.

Fixpoint check_sub (t1 t2:list_ty) { struct t1 } : bool :=
    check_sub' check_sub check_sub_map_contra t1 t2
 with check_sub_contra (t1 t2:list_ty) { struct t2 } : bool :=
    check_sub' check_sub_contra check_sub_map t1 t2
 with check_sub_map (m1 m2:contmap) { struct m1 } : bool :=
    check_map_sub' check_sub check_sub_map m1 m2
 with check_sub_map_contra (m1 m2:contmap) { struct m2 } : bool :=
    check_map_sub' check_sub_contra check_sub_map_contra m1 m2
 .

Lemma check_sub_eq' : forall n,
   (forall t1 t2, list_ty_depth t1 < n -> list_ty_depth t2 < n ->
      check_sub t1 t2 = check_sub_contra t1 t2) /\
   (forall m1 m2, contmap_depth m1 < n -> contmap_depth m2 < n ->
      check_sub_map m1 m2 = check_sub_map_contra m1 m2).

   intro n; pattern n; apply CompleteInduction; clear n; intros n Hind.
   destruct n; [ split; intros; inversion H | ].
   split; intros.

   destruct (Hind n); auto.
   destruct t1; destruct t2; simpl in *; auto.
   apply H1; auto with arith.
   apply H1; auto with arith.
   apply H1; auto with arith.
   symmetry.
   apply H2; auto with arith.
   
   destruct (Hind n); auto.
   destruct m1; destruct m2;
     simpl in *; auto;
     destruct o; destruct o0;
     simpl in *; auto.

   replace (check_sub_contra l l0) with (check_sub l l0).
   replace (check_sub_map_contra m1 m2) with (check_sub_map m1 m2).
   auto.
   apply H2; omega.
   apply H1; omega.
   apply H2; omega.
   apply H2; omega.
  Qed.

Lemma check_sub_eq : 
   forall t1 t2, check_sub t1 t2 = check_sub_contra t1 t2.

   intros.
   destruct (check_sub_eq' (1 + list_ty_depth t1 + list_ty_depth t2)).
   apply H; omega.
  Qed.

Lemma check_sub_map_eq : 
   forall m1 m2, check_sub_map m1 m2 = check_sub_map_contra m1 m2.

   intros.
   destruct (check_sub_eq' (1 + contmap_depth m1 + contmap_depth m2)).
   apply H0; omega.
  Qed.

Lemma check_sub_fixpoint : forall t1 t2,
   check_sub t1 t2 = check_sub' check_sub check_sub_map t1 t2.
 
   intros; destruct t1; simpl; auto.
   destruct t2; simpl; auto.
   rewrite check_sub_map_eq.
   auto.
  Qed.

Lemma check_sub_and_map_sound : forall p n,
  (forall t1 t2, list_ty_depth t1 < n -> list_ty_depth t2 < n ->
     check_sub t1 t2 = true -> |-- [[ t1 |p] >=> [[ t2 |p]) /\
  (forall m1 m2 x, contmap_depth m1 < n -> contmap_depth m2 < n ->
     check_sub_map m1 m2 = true ->
     |-- interp_cont p m1 x >=> interp_cont p m2 x).

  intros p n; pattern n; apply CompleteInduction; clear n; intros n Hind.
  destruct n; [ split; intros; inversion H | ].
  split; intros.

  destruct (Hind n); auto.
  rewrite check_sub_fixpoint in H1.
  destruct t1; destruct t2; simpl in *;
    try discriminate; try (apply top_sub); try (apply bot_sub);
    try (apply sub_refl).

  apply subtype_nil.
  apply_d subtype_list.
  apply H2; auto; omega.
  apply_d subtype_listmixed.
  apply H2; auto; omega.
  apply_d subtype_listcons.
  apply H2; auto; omega.
  apply_d cont_sub.
  apply H3; auto; omega.

  destruct (Hind n); auto.
  destruct m1; destruct m2;
      try destruct o; try destruct o0; simpl in *;
      try discriminate; try (apply top_sub); try (apply bot_sub);
      try (apply sub_refl).
  
  apply emptymap_top; auto.
  destruct (andb_prop _ _ H1).
  cut_with (interp_cont p m1 (x+1) >=> interp_cont p m2 (x+1)).
  cut_with (t_slot x [[ l |p] >=> t_slot x [[ l0 |p]).
  apply inter_sub.
  apply_d slot_sub.
  apply derives_axiom.
  apply H2; auto; omega.
  apply H3; auto; omega.
  cut_with (interp_cont p m1 (x+1) >=> interp_cont p m2 (x+1)).
  unfold t_sub; do 2 coalesce_modal.
  elim_at 2 implies_elim; assumed.
  apply H3; auto; omega.
  apply H3; auto; omega.
 Qed.  

Theorem check_sub_sound : forall p t1 t2,
   check_sub t1 t2 = true ->
   |-- [[ t1 |p] >=> [[ t2 |p].

  intros.
  destruct (check_sub_and_map_sound p (1 + list_ty_depth t1 + list_ty_depth t2)).
  apply H0; auto; omega.
 Qed.

Theorem check_sub_map_sound : forall p m1 m2 x,
   check_sub_map m1 m2 = true ->
   |-- interp_cont p m1 x >=> interp_cont p m2 x.

  intros.
  destruct (check_sub_and_map_sound p (1 + contmap_depth m1 + contmap_depth m2)).
  apply H1; auto; omega.
 Qed.


Fixpoint lub (s t:list_ty) { struct s } : list_ty :=
  match s, t with
  | _, lty_bot => s
  | lty_bot, _ => t

  | lty_nil, lty_nil => lty_nil
  | lty_nil, lty_list t1 => (lty_list t1)
  | lty_nil, lty_listcons t1 => (lty_list t1)

  | lty_list s1, lty_nil => (lty_list s1)
  | lty_list s1, lty_list t1 => lty_list (lub s1 t1)
  | lty_list s1, lty_listcons t1 => lty_list (lub s1 t1)

  | lty_listcons s1, lty_nil => lty_list s1
  | lty_listcons s1, lty_list t1 => lty_list (lub s1 t1)
  | lty_listcons s1, lty_listcons t1 => lty_listcons (lub s1 t1)

  | lty_cont m1, lty_cont m2 => lty_cont (mapglb m1 m2)

  | _, _ => lty_top
  end

 with glb (s t:list_ty) { struct s } : list_ty :=
  match s, t with
  | lty_top, _ => t
  | _, lty_top => s

  | lty_nil, lty_nil => lty_nil
  | lty_nil, lty_list _ => lty_nil

  | lty_list s1, lty_nil => lty_nil
  | lty_list s1, lty_list t1 => lty_list (glb s1 t1)
  | lty_list s1, lty_listcons t1 => lty_listcons (glb s1 t1)

  | lty_listcons s1, lty_list t1 => lty_listcons (glb s1 t1)
  | lty_listcons s1, lty_listcons t1 => lty_listcons (glb s1 t1)

  | lty_cont m1, lty_cont m2 => lty_cont (maplub m1 m2)

  | _, _ => lty_bot
  end

 with mapglb (m1 m2:contmap) { struct m1 } : contmap :=
   match m1, m2 with
   | cm_nil, _ => m2
   | _, cm_nil => m1

   | cm_cons None      m1'
   , cm_cons None      m2' => cm_cons None (mapglb m1' m2')

   | cm_cons (Some t1) m1'
   , cm_cons None      m2' => cm_cons (Some t1) (mapglb m1' m2')

   | cm_cons None      m1'
   , cm_cons (Some t2) m2' => cm_cons (Some t2) (mapglb m1' m2')

   | cm_cons (Some t1) m1'
   , cm_cons (Some t2) m2' => cm_cons (Some (glb t1 t2)) (mapglb m1' m2')
   end

 with maplub (m1 m2:contmap) { struct m1 } : contmap :=
   match m1, m2 with
   | cm_nil, _ => cm_nil
   | _, cm_nil => cm_nil
   | cm_cons None m1', cm_cons _ m2' => cm_cons None (maplub m1' m2')
   | cm_cons _ m1', cm_cons None m2' => cm_cons None (maplub m1' m2')
   | cm_cons (Some x) m1',
     cm_cons (Some y) m2'            => cm_cons (Some (lub x y)) (maplub m1' m2')
   end.

Lemma lub_glb_sound : forall p n,
  (forall t1 t2, list_ty_depth t1 < n ->
        |-- [[ t1 |p] >=> [[ lub t1 t2 |p] /\
        |-- [[ glb t1 t2 |p] >=> [[ t1 |p] /\
        lub t1 t2 = lub t2 t1 /\
        glb t1 t2 = glb t2 t1) /\
  (forall m1 m2 x, contmap_depth m1 < n ->
        |-- interp_cont p m1 x >=> interp_cont p (maplub m1 m2) x /\
        |-- interp_cont p (mapglb m1 m2) x >=> interp_cont p m1 x /\
        maplub m1 m2 = maplub m2 m1 /\
        mapglb m1 m2 = mapglb m2 m1).

  intros p n; pattern n; apply CompleteInduction; clear n; intros n Hind.
  destruct n.
  intuition; inversion H.
  split; intros.
  intuition.

  destruct t1; destruct t2; simpl; intros; simpl in *;
    try apply top_sub; try apply bot_sub; try apply sub_refl.
  apply subtype_nil.
  apply subtype_nil.
  apply_d subtype_list.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); auto; omega.
  apply_d subtype_list.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); auto; omega.
  apply_d subtype_listmixed.
  apply sub_refl.
  apply_d subtype_listmixed.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); auto; omega.
  apply_d subtype_listcons.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); auto; omega.
  apply_d cont_sub.
  destruct (Hind n); auto.
  destruct (H1 c c0 1%positive); intuition; omega.
    
  destruct t1; destruct t2; simpl; intros; simpl in *;
    try apply top_sub; try apply bot_sub; try apply sub_refl.
  apply subtype_nil.
  apply_d subtype_list.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); intuition; omega.  
  apply_d subtype_listmixed.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); intuition; omega.  
  apply_d subtype_listcons.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); intuition; omega.  
  apply_d subtype_listcons.
  destruct (Hind n); auto.
  destruct (H0 t1 t2); intuition; omega.  
  apply_d cont_sub.
  destruct (Hind n); auto.
  destruct (H1 c c0 1%positive); intuition; omega.  
  
  destruct t1; destruct t2; simpl; intros; simpl in *; auto;
    f_equal; destruct (Hind n); auto;
    try (destruct (H0 t1 t2); intuition; omega).
  destruct (H1 c c0 1%positive); intuition; omega.

  destruct t1; destruct t2; simpl; intros; simpl in *; auto;
    f_equal; destruct (Hind n); auto;
    try (destruct (H0 t1 t2); intuition; omega).
  destruct (H1 c c0 1%positive); intuition; omega.
  
  intuition.
  
  destruct m1; destruct m2; try destruct o; try destruct o0; simpl in *;
    try apply top_sub; try apply bot_sub; try apply sub_refl.
  apply sub_inter_right.
  apply sub_inter_left1.
  apply_d slot_sub.
  destruct (Hind n); auto.
  destruct (H0 l l0); intuition; omega.  
  apply sub_inter_left2.
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  apply sub_inter_left2.
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  

  destruct m1; destruct m2; try destruct o; try destruct o0; simpl in *;
    try apply top_sub; try apply bot_sub; try apply sub_refl.
  apply sub_inter_right.
  apply sub_inter_left1.
  apply_d slot_sub.
  destruct (Hind n); auto.
  destruct (H0 l l0); intuition; omega.
  apply sub_inter_left2.
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  apply sub_inter_right.
  apply sub_inter_left1.
  apply sub_refl.
  apply sub_inter_left2.
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  apply sub_inter_left2.
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  destruct (Hind n); auto.
  destruct (H1 m1 m2 (x+1)%positive); intuition; omega.  
  
  destruct m1; destruct m2; try destruct o; try destruct o0; simpl in *;
    f_equal; destruct (Hind n); auto;
    try (destruct (H1 m1 m2 1%positive); intuition; omega).
  f_equal; destruct (H0 l l0); intuition; omega.  
  
  destruct m1; destruct m2; try destruct o; try destruct o0; simpl in *;
    f_equal; destruct (Hind n); auto;
    try (destruct (H1 m1 m2 1%positive); intuition; omega).
  f_equal; destruct (H0 l l0); intuition; omega.  
 Qed.


Theorem lub_sound1 : forall p t1 t2,
  |-- [[ t1 |p] >=> [[ lub t1 t2 |p].

  intros.
  destruct (lub_glb_sound p (1 + list_ty_depth t1)).
  destruct (H t1 t2); intuition; omega.
 Qed.

Theorem lub_sound2 : forall p t1 t2,
  |-- [[ t2 |p] >=> [[ lub t1 t2 |p].

  intros.
  destruct (lub_glb_sound p (1 + list_ty_depth t2)).
  replace (lub t1 t2) with (lub t2 t1);
    destruct (H t2 t1); intuition; omega.
 Qed. 

Lemma lub_list_invert : forall p t1 t2 t3,
  lub (lty_list t1) t2 = lty_list t3 ->
  |-- [[ t1 |p] >=> [[ t3 |p].

  intros.
  destruct t2; simpl in *; try discriminate.
  replace t3 with t1; [ | congruence ].
  apply sub_refl.
  replace t3 with t1; [ | congruence ].
  apply sub_refl.
  simplify_eq H; intros.
  subst t3.
  apply lub_sound1.
  simplify_eq H; intros.
  subst t3.
  apply lub_sound1.
 Qed.

Definition typecheck_branch
   (pt: program_typing) (vt: contmap) (l: label): bool :=
  match pt#l with
  | None => false
  | Some vt' => check_sub_map vt vt'
  end.

Theorem typecheck_branch_sound : forall p pt vt l,
  typecheck_branch pt vt l = true ->
  check_branch p pt vt l.

  unfold typecheck_branch, check_branch; intros.
  case_eq (pt#l); intros; rewrite H0 in H; try discriminate.
  apply check_sub_map_sound.   
  auto.
 Qed.

Fixpoint typecheck_instr (pt: program_typing) (e: contmap) (i: instr)
                         {struct i}: option contmap :=
  match i with
  | i1 ;; i2 =>
      match typecheck_instr pt e i1 with
      | None    => None
      | Some e1 => typecheck_instr pt e1 i2
      end
  | instr_branch_if_nil v l =>
      match cm_get e v with
      | Some (lty_list t) =>
          if typecheck_branch pt (cm_set e v lty_nil) l
          then Some (cm_set e v (lty_listcons t))
          else None
      | Some (lty_listcons t) =>
          Some e
      | Some lty_nil =>
          if typecheck_branch pt e l
          then Some e
          else None
      | _ => None
      end
  | instr_fetch_field v1 0 v2 =>
      match cm_get e v1 with
      | Some (lty_listcons t) => Some (cm_set e v2 t)
      | _ => None
      end
  | instr_fetch_field v1 1 v2 =>
      match cm_get e v1 with
      | Some (lty_listcons t) => Some (cm_set e v2 (lty_list t))
      | _ => None
      end
  | instr_cons v1 v2 v3 =>
      match cm_get e v1, cm_get e v2 with
      | Some t1, Some t2 =>
          match lub (lty_list t1) t2 with
          | lty_list t3 => Some (cm_set e v3 (lty_listcons t3))
          | _ => None
          end
      | _, _ => None
      end
  | instr_getlabel 1%positive v => Some (cm_set e v lty_nil)
  | instr_getlabel x v =>
      match pt#x with
      | Some cont_x => Some (cm_set e v (lty_cont cont_x))
      | _ => None
      end
  | _ =>
      None
  end.


Lemma typecheck_instr_sound : forall p
  (pt: program_typing) (e e': contmap) (i:instr),

  typecheck_instr pt e i = Some e' ->
  check_instr p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1) i (interp_cont p e' 1).

  intros p pt e e' i; revert e e'; induction i;
     intros; try discriminate.

  cut (l = 1%positive \/ l <> 1%positive).
  intro Hl; destruct Hl.
  subst l; simpl in *.
  replace e' with (cm_set e v lty_nil); [ | congruence ].
  apply lm_check_instr_getnil.
  cut (Some e' =
    match pt#l with
    | Some cont_x => Some (cm_set e v (lty_cont cont_x))
    | _ => None
    end).
  intro H1; rewrite H1 in H.
  case_eq (pt#l); intros.
  rewrite H2 in H.
  rewrite H2 in H1.
  replace e' with (cm_set e v (lty_cont c));
    [ | congruence ].
  apply lm_check_instr_getlabel; auto.
  rewrite H2 in H1.
  discriminate.
  destruct l; try discriminate; simpl in *; auto.
  contradiction H0; auto.
  destruct l; auto; right; discriminate.

  simpl in H.  
  case_eq (cm_get e v); intros; rewrite H0 in H.
  destruct l0; try discriminate.
  case_eq (typecheck_branch pt e l); intros;
     rewrite H1 in H; try discriminate.
  replace e' with e; [ | congruence ].
  apply lm_check_instr_branch_on_nil; auto.
  apply typecheck_branch_sound; auto.
  case_eq (typecheck_branch pt (cm_set e v lty_nil) l); intros;
     rewrite H1 in H; try discriminate.
  replace e' with (cm_set e v (lty_listcons l0)); [ | congruence ].
  apply lm_check_instr_branch_on_list; auto.
  apply typecheck_branch_sound; auto.
  replace e' with e; [ | congruence ].
  apply lm_check_instr_branch_on_listcons with l0; auto.
  discriminate.
  
  simpl in H.
  destruct n.
  case_eq (cm_get e v); intros; rewrite H0 in H; try discriminate.
  destruct l; try discriminate.
  replace e' with (cm_set e v0 l); [ | congruence ].
  apply lm_check_instr_fetch_0; auto.
  destruct n; try discriminate.
  case_eq (cm_get e v); intros; rewrite H0 in H; try discriminate.
  destruct l; try discriminate.
  replace e' with (cm_set e v0 (lty_list l)); [ | congruence ].
  apply lm_check_instr_fetch_1; auto.
  
  unfold typecheck_instr in H.
  case_eq (cm_get e v); intros; rewrite H0 in H; try discriminate.
  case_eq (cm_get e v0); intros; rewrite H1 in H; try discriminate.
  case_eq (lub (lty_list l) l0); intros; rewrite H2 in H; try discriminate.
  replace e' with (cm_set e v1 (lty_listcons l1)); [ | congruence ].
  apply lm_check_instr_cons with l l0; auto.
  apply lub_list_invert with l0; auto.
  rewrite <- H2.
  apply lub_sound2.

  simpl in H.
  case_eq (typecheck_instr pt e i1); intros; rewrite H0 in H; try discriminate.
  apply check_instr_seq with (interp_cont p c 1).
  apply lm_pt_every.
  apply lm_ok_nec_every.
  red.
  apply ignores_incl with (Sel_I::Sel_V::nil).
  red; simpl; intuition.
  destruct (interp_ignore_nec p (1 + contmap_depth e)).
  destruct (H1 e 1%positive); auto.
  destruct (interp_ignore_nec p (1 + contmap_depth e)).
  destruct (H1 e 1%positive); auto.
  apply IHi1; auto.
  apply IHi2; auto.
 Qed.

Fixpoint typecheck_block (pt: program_typing) (e: contmap) (i: instr)
                         {struct i}: bool :=
  match i with
  | instr_halt => true
  | i1 ;; i2 =>
      match typecheck_instr pt e i1 with
      | None => false
      | Some e1 => typecheck_block pt e1 i2
      end
  | instr_jump v =>
      match cm_get e v with
      | Some (lty_cont v_cont) => check_sub_map e v_cont
      | _ => false
      end
  | _ => false
  end.


Lemma typecheck_block_sound : forall p
  (pt: program_typing) (e: contmap) (i:instr),

  typecheck_block pt e i = true ->
  check_block p (lm_pt p pt) (lm_ok p pt) (interp_cont p e 1) i.

  intros p pt e i; revert e; induction i; simpl; intros;
     try discriminate.
  case_eq (cm_get e v); intros t Hev.
  rewrite Hev in H.
  destruct t; intros; simpl in *; try discriminate.

  apply lm_check_block_jump with c; auto.
  apply check_sub_map_sound; auto.

  rewrite t in H; discriminate.

  apply check_block_halt.

  case_eq (typecheck_instr pt e i1); intros;
    rewrite H0 in H; try discriminate.
  apply check_block_seq with (interp_cont p c 1).
  apply typecheck_instr_sound; auto.
  apply IHi2; auto.
 Qed.

Definition check_one_block (pt:program_typing) (i:instr) (oe:option contmap) :=
   match oe with
   | Some e => typecheck_block pt e i
   | None   => false
   end.

Definition typecheck_blocks (p:program) (pt:program_typing) :=
 map_forall2 _ _ (check_one_block pt) p pt.

Lemma typecheck_blocks_sound : forall
   (pt:program_typing) (p:program),

   typecheck_blocks p pt = true ->
   check_blocks (interp_pt p pt) p.
   
   unfold typecheck_blocks; intros.
   unfold check_blocks; intros l i.
   unfold t_labelblock; intros.
   cut (p#l = Some i); intros.
   generalize (map_forall2_correct _ _  _ _ _ l i H H1).
   case_eq (pt#l); intros.
   exists (interp_cont p c 1); split.
   unfold interp_pt.
   change (interp_cont p c 1) with ((fun x => interp_cont p x 1) c).
   apply fmap_eqn; auto.
   refine (typecheck_block_sound _ _ _ _ _); auto.
   discriminate.
   refine (H0 sample_world _).
   simpl; intros; contradiction.
  Qed.


Lemma typecheck_blocks_safe : forall
  (pt: program_typing) (p:program) (e:contmap),

  typecheck_blocks p pt = true ->
  pt#entry_label = Some e ->
  emptymap e = true ->

  t_loaded (interp_pt p pt) p |-- t_safe p.

  intros.  
  apply program_safety with (interp_cont p e 1).
  apply typecheck_blocks_sound; auto.
  generalize H0; clear.
  destruct pt; simpl; intuition; subst; [ discriminate | auto ].
  generalize H1; clear.
  generalize (1%positive).
  induction e; simpl; intuition.
  destruct o; try discriminate.
  apply IHe; auto.
 Qed.


Fixpoint check_pt_domain (p:program) (pt:program_typing) (f:positive -> positive) { struct pt } : bool :=
  match pt with
  | Leaf => true
  | Node l o r => 
      andb (check_pt_domain p l (fun x => f (xO x)))
     (andb (check_pt_domain p r (fun x => f (xI x)))
           (match p#(f xH) with
           | Some _ => true
           | None   => false
           end))
  end.

Lemma check_pt_domain_sound : forall p pt,
   check_pt_domain p pt (fun x => x) = true ->
   |-- pt_domain_ok p (interp_pt p pt).

  intros.
  unfold pt_domain_ok.
  apply forall_intro; intro l.
  simpl_derives.
  cut (exists i:instr, p#l = Some i).
  intros [i Hi].
  apply exist_intro with i.
  apply const_intro; auto.
  revert H.
  set (f := fun x:positive => x).
  set (l' := l).
  cut (l' = f l); [ | auto ].
  clearbody l' f.
  revert p l l' x f H0.
  unfold interp_pt.
  induction pt; simpl; intros.
  destruct l; simpl in H0; discriminate.

  destruct l; simpl in *.
  apply IHpt2 with l x (fun x => f (xI x)); auto.
  elim (andb_prop _ _ H1); intros.
  elim (andb_prop _ _ H3); intros; auto.  
  apply IHpt1 with l x (fun x => f (xO x)); auto.
  elim (andb_prop _ _ H1); intros; auto.  
  elim (andb_prop _ _ H1); intros.
  elim (andb_prop _ _ H3); intros; auto.  
  subst l'.  
  destruct (p#(f 1%positive)).
  exists i; auto.
  discriminate.
 Qed.    


Definition typecheck (p:program) (pt:program_typing) : bool :=
  (typecheck_blocks p pt &&
   check_pt_domain p pt (fun x => x) &&
   match pt#entry_label with
   | Some e => emptymap e
   | None   => false
   end) % bool.

Theorem typecheck_safe : forall
  (pt: program_typing) (p:program),

  typecheck p pt = true ->

  t_loaded (interp_pt p pt) p |-- t_safe p.

  unfold typecheck; intros.
  elim (andb_prop _ _ H); intros.
  elim (andb_prop _ _ H0); intros.
  cut (exists e,
    pt#entry_label = Some e /\
    emptymap e = true).
  intro H4; destruct H4; intuition.
  eapply typecheck_blocks_safe; auto.
  apply H5.
  auto.
  destruct pt; simpl in *; try discriminate.
  destruct o; simpl; try discriminate.
  exists c; auto.
 Qed.

Theorem typecheck_safeprog : forall
  (pt:program_typing) (p:program),

  typecheck p pt = true ->
  safe_prog p.

  intros.
  cut (t_loaded (interp_pt p pt) p |-- t_safe p);
   [ | apply typecheck_safe; auto ].
  unfold safe_prog; intros.
  red in H1.
  set (W := upd_i (upd_s sample_world s) i).
  cut (forall n, t_safe p (upd_n W n)).
  intros; apply stepstar_stepN_safe; auto.
  intros.
  cut ([ t_instr i :: nil ] |-- t_safe p).
  intros H2.
  apply H2.
  simpl; intuition; subst t'.
  compute; auto.
  apply_d H0.
  unfold t_loaded; simpl_derives.
  unfold typecheck in H.
  elim (andb_prop _ _ H); intros.
  elim (andb_prop _ _ H2); intros; auto.
  apply derives_axiom.  
  apply check_pt_domain_sound; auto.

  unfold t_atlabel.
  apply exist_intro with i.
  simpl_derives; [ | assumed ].
  unfold t_labelblock.
  apply const_intro; auto.
 Qed.

