(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.
Require Import Modal.
Require Import GoedelLoeb.
Require Import Fields.

Module AdjectiveTheory (W':WORLDS) :
  ADJECTIVE_THEORY with Module W:=W'.

  Module W:=W'.
  Module FT:=FieldLogicTheory W'.

  Export FT.

  Open Local Scope pred_scope.

  Record adjective (z:polarity) (M:modality z) : Type :=
    { adj_pred      :> pred -> pred
    ; adj_pred_spec : forall p:pred, adj_pred p = (p <=> M p)
    ; adj_refl      : reflexive _ (modal_rel M)
    }.

  Lemma mkAdjective : forall z (M:modality z),
    reflexive _ (modal_rel M) ->
    adjective z M.

    intros.
    refine (Build_adjective z M (fun p => p <=> M p) _ H).
    auto.
   Qed.

  Lemma adj_intro : forall z M (J:adjective z M) p G,
    [ G ] |-- p <=> M p ->
    [ G ] |-- J p.

    intros; rewrite adj_pred_spec; auto.
   Qed.

  Lemma adj_elim : forall z M (J:adjective z M) p G t,
    [ p <=> M p :: G ] |-- t ->
    [ J p :: G ] |-- t.

    intros; rewrite adj_pred_spec; auto.
   Qed.


  Lemma add_adj : forall z M (J:adjective z M) p G t,
    [ G ] |-- J p ->
    [ M p :: G ] |-- t ->
    [ p :: G ] |-- t.
    
    intros.
    cut_with (J p); auto.
    apply adj_elim.
    cut_with (M p); auto.
    do 2 clear_at 1.
    auto.
    apply flip_equ.
    apply equ_elim.
    auto.
   Qed.

  Lemma remove_adj : forall z M (J:adjective z M) p G t,
    [ G ] |-- J p ->
    [ p :: G ] |-- t ->
    [ M p :: G ] |-- t.

    intros.
    cut_with (J p); auto.
    apply adj_elim.
    cut_with p.
    do 2 clear_at 1.
    auto.
    apply equ_elim.
    auto.
   Qed.

  Lemma add_adj_goal : forall z M (J:adjective z M) G t,
    [ G ] |-- J t ->
    [ G ] |-- M t ->
    [ G ] |-- t.

    intros.
    cut_with (M t); auto.
    cut_with (J t); auto.
    apply adj_elim.
    apply equ_elim.
    auto.
   Qed.

  Lemma remove_adj_goal : forall z M (J:adjective z M) G t,
    [ G ] |-- J t ->
    [ G ] |-- t ->
    [ G ] |-- M t.

    intros.
    cut_with t; auto.
    cut_with (J t); auto.
    apply adj_elim.
    apply flip_equ.
    apply equ_elim.
    auto.
   Qed.

  Lemma prove_adj : forall z M (J:adjective z M) p G,
    [ G ] |-- J p ->
    [ p :: G ] |-- M p.

    intros.     
    cut_with (J p); auto.
    apply remove_adj_goal with J; auto.
   Qed.

  Definition necessary  := mkAdjective _ necM necM_refl.
  Definition inevitable := mkAdjective _ eventuallyM eventuallyM_refl.
  Definition ubiquitous := mkAdjective _ everywhereM everywhereM_refl.
  Definition respects l := mkAdjective _ (respecting l) (respecting_refl l).

  Lemma p_adjective_const : forall z M (J:adjective z M) P, is_pos z ->
    |-- J (const P).

    intros.
    rewrite adj_pred_spec.
    unfold equ; split_d; unfold sub; repeat apply modal_N; auto.
    apply implies_intro; apply modal_T; auto.
    apply (adj_refl _ _ J).
   Qed.

  Lemma n_adjective_const : forall z M (J:adjective z M) P, is_neg z ->
    |-- J (const P).
  
    intros.
    rewrite adj_pred_spec.
    unfold equ; split_d; unfold sub; repeat apply modal_N; auto.
    apply implies_intro.
    apply nmodal_T; auto.
    apply (adj_refl _ _ J).
   Qed.

  Lemma adjective_const : forall z M (J:adjective z M) P,
    |-- J (const P).
  
    intros; destruct z.
    apply p_adjective_const; auto.
    apply n_adjective_const; auto.
    apply p_adjective_const; auto.
   Qed.

  Lemma bot_spec : bot = const False.
    unfold const, bot.
    auto.
   Qed.

  Lemma top_spec : top = const True.
    unfold const, top.
    auto.
   Qed.

  Lemma adjective_bot : forall z M (J:adjective z M),
    |-- J bot.

    intros.
    rewrite bot_spec.
    apply adjective_const.
   Qed.

  Lemma adjective_impl_const : forall z M (J:adjective z M) P q, is_pos z ->
    J q |-- J (const P --> q).

    intros.
    apply adj_intro.
    apply adj_elim.
    rewrite pmodal_const_impl; auto.
    unfold equ; intuition_d; unfold sub; do 2 coalesce_modal.
    intuition_d.
    apply_at 1.
    apply_at 0.
    apply const_intro; auto.
    intuition_d.
    apply_at 2.
    apply_at 0.
    apply const_intro; auto.
   Qed.

  Lemma adjective_top : forall z M (J:adjective z M),
    |-- J top.

    intros.
    rewrite top_spec.
    apply adjective_const.
   Qed.

  Lemma p_adjective_conj : forall z M (J:adjective z M) p q, is_pos z ->
    [ J p :: J q :: nil ] |-- J (p && q).
     
    intros; repeat rewrite adj_pred_spec.
    rewrite pmodal_conj; auto.
    apply_d equ_conj; auto.
   Qed.

  Lemma n_adjective_conj : forall z M (J:adjective z M) p q, is_neg z ->
    [ J p :: J q :: nil ] |-- J (p && q).

    intros; repeat rewrite adj_pred_spec.
    unfold equ; intuition_d.
    unfold sub.
    apply derives_axiom.
    apply modal_N; auto.
    apply modal_N; auto.
    intuition_d.
    apply_d nmodal_T; auto.
    apply (adj_refl _ _ J).
    unfold sub.
    clear_at 0.
    clear_at 1.
    coalesce_modal.
    coalesce_modal.
    apply implies_intro.
    cut_with (M p && M q).
    intuition_d.
    apply_at 3; auto.
    apply_at 4; auto.
    apply_d modal_distrib_conj; auto.
   Qed.

  Lemma adjective_conj : forall z M (J:adjective z M) p q,
    [ J p :: J q :: nil ] |-- J (p && q).

    intros; destruct z.
    apply p_adjective_conj; auto.
    apply n_adjective_conj; auto.
    apply p_adjective_conj; auto.
   Qed.

  Lemma p_adjective_disj : forall z M (J:adjective z M) p q, is_pos z ->
    [ J p :: J q :: nil ] |-- J (p || q).
  
    intros.
    repeat rewrite adj_pred_spec.
    unfold equ; intuition_d.  
    clear_at 1.
    clear_at 2.
    unfold sub.
    coalesce_modal.
    coalesce_modal.
    intuition_d.
    apply_d modal_distrib_disj.
    left_d; apply_at 1; auto.
    apply_d modal_distrib_disj.
    right_d; apply_at 2; auto.
    apply derives_axiom.
    unfold sub; repeat apply modal_N; auto.
    apply implies_intro; apply modal_T; auto.
    apply (adj_refl _ _ J).
   Qed.

  Lemma n_adjective_disj : forall z M (J:adjective z M) p q, is_neg z ->
    [ J p :: J q :: nil ] |-- J (p || q).

    intros.
    repeat rewrite adj_pred_spec.
    rewrite nmodal_disj; auto.
    apply_d equ_disj; auto.
   Qed.

  Lemma adjective_disj : forall z M (J:adjective z M) p q,
    [ J p :: J q :: nil ] |-- J (p || q).

    intros; destruct z.
    apply p_adjective_disj; auto.
    apply n_adjective_disj; auto.
    apply p_adjective_disj; auto.
   Qed.

  Lemma sub_every_nec : forall p q,
    p >=> q |-- []!(p >=> q).

    intros; unfold sub.
    rewrite every_nec_commute.
    repeat rewrite modal_idem; auto.
   Qed.

  Lemma equ_every_nec : forall p q,
    p <=> q |-- []!(p <=> q).
   
    intros; unfold equ.
    repeat rewrite pmodal_conj; auto.
    split_d; apply_d sub_every_nec; auto.
   Qed.

  Lemma all_Jf_nec_every : forall z M (J:adjective z M) A (f:A->pred),
    All_ a:A, J (f a) |-- []!(All_ a:A, J (f a)).

    intros.
    repeat rewrite pmodal_forall; auto.
    intro_d; intro a.
    spec_at 0 a.
    repeat rewrite adj_pred_spec.
    apply equ_every_nec.
   Qed.

  Lemma p_adjective_forall : forall z M (J:adjective z M) A (f:A->pred), is_pos z ->
    All_ a:A, J (f a) |-- J (all f).

    intros.
    rewrite adj_pred_spec.
    cut_with ([]!(All_ a:A, J (f a)));
      [ | apply_d all_Jf_nec_every; auto ].
    apply derives_weaken_head.
    unfold equ; split_d; unfold sub; 
     do 2 apply derives_modal;
     apply implies_intro.
    rewrite pmodal_forall; auto.
    apply forall_intro; intro a.
    spec_at 0 a.
    spec_at 1 a.
    rewrite adj_pred_spec.
    unfold equ; apply conj_elim.
    clear_at 1.
    apply sub_project_herenow.
    apply_d (modal_T _ M); auto.
    apply (adj_refl _ _ J).
   Qed.

  Lemma n_adjective_forall : forall z M (J:adjective z M) A (f:A->pred), is_neg z ->
    All_ a:A, J (f a) |-- J (all f).
    
    intros.
    rewrite adj_pred_spec.
    cut_with ([]!(All_ a:A, J (f a)));
      [ | apply_d all_Jf_nec_every; auto ].
    apply derives_weaken_head.
    (in_goal unfold equ) ; split_d;
      unfold sub; do 2 apply derives_modal;
      apply implies_intro.
    apply_d nmodal_T; auto.
    apply (adj_refl _ _ J).
    apply forall_intro; intro a.
    cut_with (All_ a:A, M (f a)).
    clear_at 1.
    spec_at 0 a.
    spec_at 1 a.
    rewrite adj_pred_spec.
    unfold equ; apply conj_elim.
    clear_at 0.
    apply sub_project_herenow.
    apply_d modal_distrib_forall.
    auto.
   Qed.
   
  Lemma adjective_forall : forall z M (J:adjective z M) A (f:A->pred),
    All_ a:A, J (f a) |-- J (all f).

    intros; destruct z.
    apply p_adjective_forall; auto.
    apply n_adjective_forall; auto.
    apply p_adjective_forall; auto.
   Qed.

  Lemma p_adjective_exist : forall z M (J:adjective z M) A (f:A->pred), is_pos z ->
    All_ a:A, J (f a) |-- J (ex f).

    intros.
    rewrite adj_pred_spec.
    cut_with ([]!(All_ a:A, J (f a)));
      [ | apply_d all_Jf_nec_every; auto ].
    clear_at 1.
    unfold equ; split_d; (in_goal unfold sub);
      do 2 apply derives_modal;
      apply implies_intro.
    apply_d modal_distrib_exist.
    elim_at 0; intro a.
    apply exist_intro with a.
    spec_at 1 a.
    rewrite adj_pred_spec.
    unfold equ.
    elim_at 0.
    clear_at 1.
    apply sub_project_herenow.
    apply_d (modal_T _ M); auto.
    apply (adj_refl _ _ J).
   Qed.

  Lemma n_adjective_exist : forall z M (J:adjective z M) A (f:A->pred), is_neg z ->
    All_ a:A, J (f a) |-- J (ex f).

    intros.
    rewrite adj_pred_spec.    
    cut_with ([]!(All_ a:A, J (f a)));
      [ | apply_d all_Jf_nec_every; auto ].
    clear_at 1.
    unfold equ; split_d; (in_goal unfold sub);
      do 2 apply derives_modal;
      apply implies_intro.
    apply_d nmodal_T; auto.
    apply (adj_refl _ _ J).
    rewrite <- nmodal_exist; auto.
    elim_at 0; intro a.
    apply exist_intro with a.
    spec_at 1 a.
    rewrite adj_pred_spec.
    unfold equ.
    elim_at 0.
    clear_at 0.
    apply sub_project_herenow.
   Qed.

  Lemma adjective_exist : forall z M (J:adjective z M) A (f:A->pred),
    All_ a:A, J (f a) |-- J (ex f).

    intros; destruct z.
    apply p_adjective_exist; auto.
    apply n_adjective_exist; auto.
    apply p_adjective_exist; auto.
   Qed.

  Lemma adjective_recursive' : forall F M,
    contractive F ->
    (forall P, |> (P <=> M P) |-- F (M P) <=> M (F P)) ->
    |-- rec F <=> M (rec F).

    intros.
    apply loeb_rule.
    in_goal rewrite fold_unfold; auto.
    apply_d equ_trans; [ apply_d H | apply H0 ]; auto.
   Qed.

  Lemma adjective_recursive : forall z M (J:adjective z M) F,
    contractive F ->
    (All_ P:pred, []((|> (J P)) --> (F (M P) <=> M (F P)))) |-- J (rec F).

    intros.
    apply generalized_loeb_rule.
    apply_d nec_later.
    rewrite pmodal_forall; auto.
    intro_d; intro a.
    spec_at 0 a; auto.
    rewrite adj_pred_spec.
    in_goal rewrite fold_unfold; auto.
    apply_d equ_trans; [ apply_d H; exact_at 1 | ].
    spec_at 0 (rec F).    
    elim_at 0; auto.
    rewrite adj_pred_spec.
    apply_at 0.
    auto.
   Qed.

  Lemma adjective_modal : forall z M (J:adjective z M) p,
    transitive _ (modal_rel M) ->
    |-- J (M p).

    intros; apply adj_intro.
    rewrite modal_idem; auto.
    apply (adj_refl _ _ J).
   Qed.

  Lemma later_necessary : forall p,
    |-- necessary (|> p).

    intros; apply adj_intro.
    rewrite nec_later_later; auto.
   Qed.

  Lemma nec_necessary : forall p,
    |-- necessary ([] p).
    
    intros; apply adj_intro.
    rewrite modal_idem; auto.
   Qed.

  Lemma projecting_necessary : forall x a,
    |-- necessary (projecting x a).

    intros; apply adj_intro.
    unfold equ; split_d; unfold sub; repeat apply modal_N; auto.
    apply implies_intro.
    rewrite necM_spec.
    unfold nec, nec_rel, appPosMode, projecting, derives; simpl; intuition.
    rewrite <- H0.
    destruct H as [n H].
    rewrite (w_upd_noninterference_n n w w' x); auto.
   Qed.

  Lemma field_pred_necessary : forall x f,
    All_ a:field_type x, necessary (f a) |-- necessary (field_pred x f).

    intros; apply adj_intro.
    cut_with ([]!(All_ a:field_type x, necessary (f a)));
      [ | apply all_Jf_nec_every; auto ].
    clear_at 1.
    unfold equ; split_d; unfold sub; do 2 apply derives_modal;
      intuition_d.
    apply (with_projecting x); intro a.
    move_at 1; apply field_pred_elim with a; auto.
    spec_at 2 a.
    apply adj_elim.
    cut_with ([](f a)); [ | apply flip_equ; apply equ_elim; auto ].
    do 2 clear_at 1.
    rewrite necM_spec.
    unfold nec, nec_rel, appPosMode, projecting, derives; simpl; intuition.
    unfold field_pred in *.
    destruct H1 as [n H1].
    rewrite <- (w_upd_noninterference_n n w w' x); auto.
    rewrite H.
    apply H0.
    exists n; auto.
   Qed.

  Lemma respecting_necessary : forall l p,
    necessary p |-- necessary (respecting l p).

    intros; apply adj_intro.
    apply adj_elim.        
    rewrite <- respecting_nec_commute; auto.
    apply equ_respecting.
   Qed.

  Lemma later_respects : forall l p,
    respects l p |-- respects l (|> p).

    intros; apply adj_intro; apply adj_elim.
    rewrite respecting_later_commute; auto.
    apply equ_later.
   Qed.

  Lemma nec_respects : forall l p,
    respects l p |-- respects l ([] p).

    intros; apply adj_intro; apply adj_elim.
    rewrite respecting_nec_commute; auto.
    apply equ_nec.
   Qed.

  Lemma projecting_respects : forall x l a,
    In x l ->
    |-- respects l (projecting x a).

    intros; apply adj_intro.
    unfold equ; intuition_d; unfold sub; repeat apply modal_N; auto.
    intro_d.
    rewrite respecting_spec.
    unfold appPosMode, respecting_rel, derives, projecting; simpl; intuition.
    rewrite <- H1.
    rewrite H3; auto.
   Qed.

  Lemma field_pred_respects : forall x l f,
    In x l ->
    (All_ a:field_type x, respects l (f a)) |--
       respects l (field_pred x f).

    intros.
    apply adj_intro.
    unfold equ; intuition_d.
    2: apply derives_axiom.
    2: unfold sub; repeat apply modal_N; auto.
    cut_with ([]!(All_ a:field_type x, respects l (f a)));
     [ | apply all_Jf_nec_every; auto ].
    clear_at 1.
    unfold sub; do 2 apply derives_modal.
    intro_d.
    apply (with_projecting x); intro a.
    move_at 1; apply (field_pred_elim _ a); auto.
    spec_at 2 a.
    apply adj_elim.
    move_at 2.
    apply (add_adj _ _ (respects l)).
    apply_d projecting_respects.
    cut_with (respecting l (f a)).
    coalesce_modal.
    apply field_pred_intro with a; auto.
    clear_at 0.
    apply flip_equ; apply equ_elim.
    auto.
   Qed.

  Lemma respects_nil_ubiquitous : forall p,
    respects nil p = ubiquitous p.

    intros.
    repeat rewrite adj_pred_spec.
    rewrite everywhere_respecting_nil.
    auto.
   Qed.

  Lemma respecting_respects : forall l1 l2 p,
    incl l2 l1 ->
    |-- respects l1 (respecting l2 p).

    intros.
    apply adj_intro.
    unfold equ; split_d; unfold sub; repeat apply modal_N; auto.
    intro_d.
    apply_d (respecting_incl l2).
    apply modal_4; auto.
   Qed.

  Lemma with_subst_necessary : forall z p,
    necessary p |-- necessary (with_subst z p).

    intros.    
    apply adj_intro.
    rewrite <- with_subst_nec_commute.
    apply adj_elim.
    apply equ_with_subst.
   Qed.

  Lemma with_subst_respects : forall l z p,
    respects l p |-- respects l (with_subst z p).

    intros.
    apply adj_intro.
    apply adj_elim.
    unfold equ.    
    intuition_d.
    clear_at 1.
    cut_with ([]!!(p --> respecting l p)).
    clear_at 1.
    unfold sub; do 2 (apply derives_modal; auto).
    intro_d.
    rewrite with_subst_spec.
    rewrite respecting_spec.
    rewrite everywhere_spec.
    unfold everywhere, appPosMode, respecting_rel, with_subst_rel, implies, everywhere_rel.
    unfold derives; simpl; intuition.
    apply H with (world_update z w); auto.
    clear; induction z; simpl; auto.
    destruct a.
    rewrite <- world_ttl_noninterference.
    auto.
    split; intros.
    rewrite <- H1.
    generalize (H3 x H5).
    clear; induction z; simpl; auto.
    destruct a; simpl; intros.
    destruct (field_eq_dec x x0); subst.
    rewrite proj_upd_same.
    rewrite proj_upd_same.
    auto.
    rewrite proj_upd_diff; auto.
    rewrite proj_upd_diff; auto.
    rewrite <- H1.
    revert H4; clear; intros; induction z; simpl; auto.
    destruct a.
    rewrite <- world_ttl_noninterference.
    rewrite <- world_ttl_noninterference.
    auto.
    unfold sub.
    rewrite modal_idem; auto.
    apply derives_axiom.
    unfold sub.
    do 2 (apply modal_N; auto).
   Qed.

  Lemma adj_nec : forall z M (J:adjective z M) p,
    J p = [](J p).

    intros; apply derives_equal; auto.
    rewrite adj_pred_spec.
    unfold equ.
    rewrite pmodal_conj; auto.
    unfold sub.
    intuition_d.
    rewrite modal_idem; auto.
   Qed.

  Lemma adj_every : forall z M (J:adjective z M) p,
    J p = !(J p).

    intros; apply derives_equal; auto.
    rewrite adj_pred_spec.
    unfold equ.
    rewrite pmodal_conj; auto.
    unfold sub.
    rewrite every_nec_commute.
    rewrite modal_idem; auto.
    rewrite every_nec_commute.
    rewrite modal_idem; auto.
   Qed.

  Lemma adj_nec_every : forall z M (J:adjective z M) p,
    J p = []!(J p).

    intros; apply derives_equal; auto.
    rewrite adj_pred_spec.
    unfold equ.
    rewrite pmodal_conj; auto.
    rewrite pmodal_conj; auto.
    unfold sub.
    intuition_d.
    rewrite every_nec_commute.
    rewrite modal_idem; auto.
    rewrite modal_idem; auto.
    rewrite every_nec_commute.
    rewrite modal_idem; auto.
   Qed.
End AdjectiveTheory.
