(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.
Require Import Worlds.

Module LogicTheory (W':WORLDS) : LOGIC_THEORY with Module W:=W'.
  Module W := W'.
  Module WT := WorldTheory W'.

  Export WT.

  Definition pred := world -> Prop.

  Fixpoint all_true (l:list pred) (w:world) { struct l } : Prop :=
    match l with
    | nil => True
    | x :: l' => x w /\ all_true l' w
    end.
    
  Definition derives (l:list pred) (p:pred) := 
      forall w, all_true l w -> p w.

  Delimit Scope pred_scope with pred.

  Notation "|--  t" := (derives nil t) (at level 80, no associativity) : pred_scope.
  Notation " t1  |--  t2" := (derives (t1::nil) t2) (at level 80, no associativity) : pred_scope.
  Notation " [  G  ]  |--  t" := (derives G t)
        (at level 80, no associativity) : pred_scope.

  (* Primitives *)
  Definition bot: pred := fun _ => False.
  Definition top: pred := fun _ => True.
  Definition disj (t1 t2: pred) : pred := fun w => t1 w \/ t2 w.
  Definition conj (t1 t2: pred) : pred := fun w => t1 w /\ t2 w.
  Definition implies (t1 t2: pred) : pred := fun w => t1 w -> t2 w.
  Definition ex (A: Type) (F: A -> pred) := fun w => exists x:A, F x w.
  Implicit Arguments ex.
  Definition all (A: Type) (F: A -> pred) := fun w => forall x:A, F x w.
  Implicit Arguments all.
  Definition const (P:Prop) : pred := fun _ => P.
  Definition neg (p:pred) := implies p bot.
  Definition projecting (x:field) (a:field_type x) : pred :=
     fun w => w_proj x w = a.
  Definition field_pred (x:field) (P:field_type x -> pred) : pred :=
     fun w => P (w_proj x w) w.

  Notation "'Ex_'  x ':' T ',' P " := (ex (fun x:T => P%pred)) (at level 65, x at level 99) : pred_scope.
  Notation "'All_'  x ':' T  ',' P " := (all (fun x:T => P%pred)) (at level 65, x at level 99) : pred_scope.
  Infix "||" := disj (at level 50, left associativity) : pred_scope.
  Infix "&&" := conj (at level 40, left associativity) : pred_scope.
  Notation "P '-->' Q" := (implies P Q) (at level 60, right associativity) : pred_scope.

  Open Local Scope pred_scope.

  Tactic Notation "in_goal" tactic(tac) :=
    let Z := fresh "Z" in
      match goal with [ |- derives ?G ?t ] =>
        set (Z:=G); tac; subst Z
      end.

  Lemma derives_equal : forall t1 t2,
    t1 |-- t2 ->
    t2 |-- t1 ->
    t1 = t2.
    
    intros.
    refine (extensionality _ _ _); intro x; apply prop_ext.
    split; intros.
    apply H; simpl; auto.
    apply H0; simpl; auto.    
   Qed.

  Lemma equal_derives : forall t1 t2,
    t1 = t2 -> t1 |-- t2.

    intros; rewrite H; red; simpl; intuition.
   Qed.

  Lemma eta_equal : forall t:pred,
    t = (fun w:world => t w).

    intros.
    refine (extensionality _ _ _).
    auto.
   Qed.

  (* Structural rules for derives judgements *)

  Lemma derives_weaken : forall G1 G2 t,
    (forall P, In P G1 -> In P G2) ->
    [ G1 ] |-- t ->
    [ G2 ] |-- t.

    unfold derives in *; intros; simpl in *; intuition.
    apply H0; intros.
    clear H0.
    induction G1; simpl in *; intuition.
    clear IHG1.
    cut (In a G2).
    clear H.
    induction G2; simpl in *; intuition.
    subst a0; auto.
    apply H; auto.
   Qed.

  Lemma derives_weaken_head : forall P G t,
    [ P :: nil ] |-- t ->
    [ P :: G   ] |-- t.

    intros.
    apply derives_weaken with (P::nil); simpl; intros; intuition.
  Qed.

  Ltac weaken_to l
    := refine (derives_weaken l _ _ _ _);
      try (simpl; intuition; fail).

  Ltac weaken_with Z
    := refine (derives_weaken _ _ _ _ _);
      [ | apply Z ]; try (simpl; intuition; fail).

  Lemma derives_axiom : forall G t,
    |-- t ->
    [ G ] |-- t.

    intros l t; apply derives_weaken; simpl; intros; contradiction.
  Qed.

  Lemma derives_permute : forall G1 G2 t,
    Permutation G1 G2 ->
    [ G1 ] |-- t ->
    [ G2 ] |-- t.

    intros l1 l2 t H.
    apply derives_weaken.
    intro P.  
    apply Permutation_in.
    auto.
  Qed.

  Lemma derives_move_left : forall P l1 l2 t,
    [ P :: l1 ++ l2 ] |-- t ->
    [ l1 ++ P :: l2 ] |-- t.

    intros P l1 l2 t; apply derives_permute.
    apply Permutation_trans with ((P :: l2) ++ l1).
    simpl.
    apply perm_skip.
    apply Permutation_app_swap.
    apply Permutation_app_swap.
  Qed.

  Lemma derives_dup_left : forall P l1 l2 t,
    [ P :: l1 ++ P :: l2 ] |-- t ->
    [ l1 ++ P :: l2 ] |-- t.

    intros P l1 l2 t.
    apply derives_weaken; intros.
    simpl in H; destruct H; auto.
    subst P; apply in_or_app; simpl; auto.
  Qed.

  Lemma derives_assumption : forall G t,
    In t G ->
    [ G ] |-- t.

    unfold derives; intros; auto.
    induction G; simpl in *; intuition.
    subst; auto.
  Qed.

  Lemma assumed1 : forall t G,
    [ t :: G ] |-- t.
    
    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed2 : forall t1 t G,
    [ t1 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed3 : forall t1 t2 t G,
    [ t1 :: t2 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed4 : forall t1 t2 t3 t G,
    [ t1 :: t2 :: t3 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed5 : forall t1 t2 t3 t4 t G,
    [ t1 :: t2 :: t3 :: t4 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed6 : forall t1 t2 t3 t4 t5 t G,
    [ t1 :: t2 :: t3 :: t4 :: t5 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed7 : forall t1 t2 t3 t4 t5 t6 t G,
    [ t1 :: t2 :: t3 :: t4 :: t5 :: t6 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Lemma assumed8 : forall t1 t2 t3 t4 t5 t6 t7 t G,
    [ t1 :: t2 :: t3 :: t4 :: t5 :: t6 :: t7 :: t :: G ] |-- t.

    intros; apply derives_assumption; simpl; intuition.
   Qed.

  Hint Resolve assumed1 assumed2 assumed3 assumed4
     assumed5 assumed6 assumed7 assumed8.

  Ltac assumed := auto; apply derives_assumption; simpl; intuition; fail.

  Lemma derives_left : forall l t,
    [ t :: l ] |-- t.

    intros; apply derives_assumption; simpl; auto.
  Qed.

  Lemma derives_cut : forall c t G,
    [ c :: G ] |-- t ->
    [ G ]      |-- c ->
    [ G ]      |-- t.

    unfold derives; intros.
    apply H; simpl; intuition.
   Qed.

  Ltac cut_with t := apply derives_cut with t.

  Lemma derives_apply : forall G G' t,
    (fold_right (fun p Z => [ G ] |-- p /\ Z) True G') ->
    [ G' ] |-- t ->
    [ G ]  |-- t.

    intros.
    revert t G H H0.
    induction G'; simpl in *; intuition.
    apply derives_axiom; auto.
    generalize (IHG' (implies a t) G H2).
    intros.
    clear IHG' H2.
    unfold derives in *.        
    simpl in *.
    intros.
    apply H; intuition.
    red; intros.
    apply H0; intuition.
   Qed.

  Ltac split_right := try trivial; try (split; [ | trivial; split_right ]).

  Ltac apply_d H := refine (derives_apply _ _ _ _ _);
    [ | apply H; auto with modal ]; simpl; split_right.

  Ltac cut_using X :=
   match type of X with
     derives ?G ?t  =>
     cut_with t; [ | apply_d X ]
   end.

  Lemma in_all_true : forall G1 G2,
     (forall x, In x G2 -> In x G1) ->
     (forall w, all_true G1 w -> all_true G2 w).

     intros.
     induction G2; simpl in *; intuition.
     cut (In a G1).
     clear H IHG2.
     induction G1; simpl in *; intuition.
     subst a0; auto.
     apply H; auto.
    Qed.


  Fixpoint split_list (n:nat) (l1 l2:list pred) {struct n}
    : (option pred * list pred * list pred) :=

    match n, l2 with
      | 0   , x :: l2' => (Some x, rev l1, l2')
      | S n', x :: l2' => split_list n' (x::l1) l2'
      | _   , nil      => (None, rev l1, nil)
    end.
  
  Lemma split_list_together : forall n l l' z l1 l2,
    split_list n l' l = (z, l1, l2) ->
    rev l' ++ l = l1 ++ (match z with Some x => x::nil | None => nil end) ++ l2.
    
    intro n; induction n; destruct l; simpl; intros;
      injection H; intros; subst; simpl; auto.
    replace (rev l' ++ p :: l) with (rev (p::l') ++ l);
    [ | simpl; rewrite app_ass; auto ].
    apply IHn; auto.
  Qed.

  Definition calc_move_at (n:nat) (l:list pred) :=
    match split_list n nil l with
      | ( None, _, _ )    => l
      | ( Some x, l1, l2) => x :: l1 ++ l2
    end.

  Definition calc_dup_at (n:nat) (l:list pred) :=
    match split_list n nil l with
      | ( None, _, _ )    => l
      | ( Some x, l1, l2) => x :: l1 ++ x :: l2
    end.
  
  Lemma calc_move_at_in : forall l n x,
    In x (calc_move_at n l) -> In x l.

    intros l n x; unfold calc_move_at.
    case_eq (split_list n nil l); intros;
        destruct p; destruct o; auto.

    change l with (rev nil ++ l).
    rewrite (split_list_together n l nil (Some p) l1 l0); auto.
    apply in_or_app.
    simpl in H0; destruct H0.
    subst; right; simpl; auto.
    elim (in_app_or _ _ _ H0); intros; simpl; auto.
   Qed.

  Lemma calc_dup_at_in : forall l n x,
    In x (calc_dup_at n l) -> In x l.

    intros l n x; unfold calc_dup_at;
      case_eq (split_list n nil l); intros;
        destruct p; destruct o; auto.

    change l with (rev nil ++ l).
    rewrite (split_list_together n l nil (Some p) l1 l0); auto.
    simpl in H0; destruct H0; auto.
    subst p.
    apply in_or_app.
    right; simpl; auto.
   Qed.  

  Ltac move_at n :=
    match goal with
      | [ |- derives ?G ?t ] =>
        cut (derives (calc_move_at n G) t);
          [ apply derives_weaken; apply calc_move_at_in
          | unfold calc_move_at; simpl
          ]
    end.

  Ltac dup_at n :=
    match goal with
      | [ |- derives ?G ?t ] =>
        cut (derives (calc_dup_at n G) t);
          [ apply derives_weaken; apply calc_dup_at_in
          | unfold calc_dup_at; simpl
          ]
      end.

(*
  Ltac elim_at n A :=
       match n with
       | O => apply A
       | _ => move_at n; apply A
       end.

  Ltac elim_dup_at n A := dup_at n; apply A.
*)

  Ltac exact_at n := move_at n; apply derives_left.

  Lemma derives_drop_head : forall t' t G,
    [ G ] |-- t ->
    [ t' :: G ] |-- t.
    
    intros; apply derives_weaken with G; simpl; intuition.
  Qed.

  Ltac clear_at x := move_at x; apply derives_drop_head.


  (* introduction and elimination rules for
     the basic logical constructs *)



  Lemma bot_elim : forall t G,
    [ bot :: G ] |-- t.

    unfold bot, derives; intros; simpl in *.
    intuition.
   Qed.

  Lemma top_intro : forall G,
    [ G ] |-- top.
    
    do 2 (red; intros); auto.
  Qed.

  Lemma disj_intro1 : forall t1 t2 G,
    [ G ] |-- t1 ->
    [ G ] |-- (t1 || t2).

    red; unfold disj; intros.
    simpl in *; intuition.
  Qed.  

  Lemma disj_intro2 : forall t1 t2 G,
    [ G ] |-- t2 ->
    [ G ] |-- (t1 || t2).

    red; unfold disj; intros.
    simpl in *; intuition.
   Qed.

  Lemma disj_elim : forall t1 t2 t G,
    [ t1 :: G ] |-- t ->
    [ t2 :: G ] |-- t ->
    [ t1 || t2 :: G ] |-- t.

    unfold derives; intros.
    simpl in *.
    intuition.
    destruct H2.
    apply H; intuition.
    apply H0; intuition.
   Qed.

  Lemma derives_disj_commute : forall t1 t2,
    t1 || t2 |-- t2 || t1.

    intros; apply disj_elim.
    apply disj_intro2; assumed.
    apply disj_intro1; assumed.
   Qed.  

  Lemma disj_commute : forall t1 t2,
    t1 || t2 = t2 || t1.

    intros; apply derives_equal; apply derives_disj_commute.
  Qed.

  Lemma disj_assoc : forall t1 t2 t3,
    t1 || (t2 || t3) = (t1 || t2) || t3.

    intros; apply derives_equal.
    apply disj_elim.
    apply disj_intro1.
    apply disj_intro1; assumed.
    apply disj_elim.
    apply disj_intro1.
    apply disj_intro2; assumed.
    apply disj_intro2; assumed.

    apply disj_elim.
    apply disj_elim.
    apply disj_intro1; assumed.
    apply disj_intro2.
    apply disj_intro1; assumed.
    apply disj_intro2.
    apply disj_intro2; assumed.
  Qed.

  Lemma disj_bot_unit : forall t,
    bot || t = t.

    intro t; apply derives_equal.
    apply disj_elim.
    apply bot_elim.
    assumed.
    apply disj_intro2; assumed.
  Qed.

  Lemma conj_intro : forall t1 t2 G,
    [ G ] |-- t1 ->
    [ G ] |-- t2 ->
    [ G ] |-- (t1 && t2).

    unfold derives; intros; split.
    apply H; auto.
    apply H0; auto.
  Qed.

  Lemma conj_elim : forall t1 t2 t G,
    [ t1 :: t2 :: G ] |-- t ->
    [ t1 && t2 :: G ] |-- t.

    unfold derives; intros.
    apply H; intros.
    simpl in H0; intuition.
    destruct H1.
    simpl; intuition.
   Qed.

  Lemma conj_elim1 : forall t1 t2 t G,
    [ t1 :: G ]       |-- t ->
    [ t1 && t2 :: G ] |-- t.

    intros; apply conj_elim; weaken_with H.
  Qed.

  Lemma conj_elim2 : forall t1 t2 t G,
    [ t2 :: G ]       |-- t ->
    [ t1 && t2 :: G ] |-- t.

    intros; apply conj_elim; weaken_with H.
  Qed.

  Lemma derives_conj_commute : forall t1 t2,
    t1 && t2 |-- t2 && t1.

    intros; apply conj_elim; apply conj_intro; assumed.
  Qed.

  Lemma conj_commute : forall t1 t2,
    t1 && t2 = t2 && t1.

    intros; apply derives_equal; apply derives_conj_commute.
  Qed.
  
  Lemma conj_assoc : forall t1 t2 t3,
    t1 && (t2 && t3) = (t1 && t2) && t3.

    intros; apply derives_equal.
    apply conj_elim; move_at 1; apply conj_elim.
    apply conj_intro.
    apply conj_intro.
    assumed.  
    assumed.  
    assumed.  
    apply conj_elim; apply conj_elim.
    apply conj_intro.
    assumed.
    apply conj_intro.
    assumed.
    assumed.
  Qed.

  Lemma intersection_top_unit : forall t,
    top && t = t.

    intros; apply derives_equal.
    apply conj_elim; assumed.
    apply conj_intro.
    apply top_intro.
    assumed.
  Qed.

  Lemma conj_bot : forall t,
    bot && t = bot.

    intro t; apply derives_equal.
    apply conj_elim.
    assumed.
    apply bot_elim.
  Qed.

  Lemma disj_top : forall t,
    top || t = top.

    intro t; apply derives_equal.
    apply top_intro.
    apply disj_intro1.
    assumed.
  Qed.

  Lemma distrib_disj_conj : forall n m p,
    (n || m) && p = (n&&p) || (m&&p).

    intros; apply derives_equal.
    apply conj_elim.
    apply disj_elim.
    apply disj_intro1; apply conj_intro; assumed.
    apply disj_intro2; apply conj_intro; assumed.
    apply disj_elim; apply conj_elim;
    apply conj_intro; try assumed.
    apply disj_intro1; assumed.
    apply disj_intro2; assumed.
  Qed.

  Lemma distrib_conj_disj : forall n m p,
    (n&&m) || p = (n||p) && (m||p).

    intros; apply derives_equal.
    apply conj_intro.
    apply disj_elim.
    apply conj_elim.
    apply disj_intro1; assumed.
    apply disj_intro2; assumed.
    apply disj_elim.
    apply conj_elim.
    apply disj_intro1; assumed.
    apply disj_intro2; assumed.
    apply conj_elim.
    apply disj_elim.
    move_at 1; apply disj_elim.
    apply disj_intro1; apply conj_intro; assumed.
    apply disj_intro2; assumed.
    apply disj_intro2; assumed.
  Qed.

  Lemma implies_intro : forall t1 t2 G,
    [ t1 :: G ] |-- t2 ->
    [ G ]       |-- (t1 --> t2).

    unfold derives; red; intros.
    apply H; intros.
    simpl; intuition.
   Qed.

  Lemma implies_elim : forall t1 t2 G,
    [ G ] |-- t1 ->
    [ (t1 --> t2) :: G ] |-- t2.

    unfold derives; intros.
    cut ((t1 --> t2) w).
    intros H1; apply H1.
    apply H; intros.
    simpl in *; intuition.
    simpl in *; intuition.
   Qed.

  Lemma modus_ponens : forall p q G t,
    [ G ] |-- p ->
    [ q :: G ] |-- t ->
    [ (p --> q) :: G ] |-- t.

    intros.
    cut_with q.
    weaken_with H0.
    apply implies_elim.
    auto.
   Qed.

  Lemma exist_intro : forall A (f:A -> pred) (x:A) G,
    [ G ] |-- (f x) ->
    [ G ] |-- (ex f).

    unfold derives; intros.
    red; exists x.
    apply H; auto.
   Qed.

  Lemma exist_elim : forall A (f:A -> pred) t G,
    (forall x:A, [ f x :: G ] |-- t) ->
    [ ex f :: G ] |-- t.

    unfold derives; intros.
    simpl in H0; intuition.
    destruct  H1.
    apply (H x); intros.
    simpl; intuition.
   Qed.

  Lemma forall_intro : forall A (f:A -> pred) G,
    (forall x:A, [  G ] |-- f x) ->
    [ G ] |-- all f.

    unfold derives; red; intros.
    apply H; auto.
   Qed.

  Lemma forall_elim : forall A (x:A) (f:A -> pred) t G,
    [ f x :: G ] |-- t ->
    [ all f :: G ] |-- t.

    unfold derives; intros.
    apply H; intros.
    simpl in *; intuition.
   Qed.

  Lemma const_intro : forall G (P:Prop),
    P -> 
    [ G ] |-- const P.

    intros G P H; unfold derives, const; intros; auto.
   Qed.

  Lemma const_elim : forall t G (P:Prop),
    (P -> [ G ] |-- t) ->
    [ const P :: G ] |-- t.

    intros t G P H.
    unfold derives, const; intros.
    simpl in *; intuition.
   Qed.

  Lemma with_projecting : forall x t G,
    (forall a:field_type x, [ projecting x a :: G ] |-- t) ->
    [ G ] |-- t.

    unfold derives; intuition.
    apply (H (w_proj x w)).
    simpl; intuition.
    red; auto.
   Qed.

  Lemma field_pred_elim : forall x a f t G,
    [ G ] |-- projecting x a ->
    [ f a :: G ] |-- t ->
    [ field_pred x f :: G ] |-- t.

    intros.
    cut_with (f a).
    weaken_with H0.
    unfold derives in *; simpl in *; intuition.
    red in H2.
    cut (projecting x a w).
    unfold projecting; intros.
    subst a; auto.
    apply H; auto.
   Qed.

  Lemma field_pred_intro : forall x a f G,
    [ G ] |-- projecting x a ->
    [ G ] |-- f a ->
    [ G ] |-- field_pred x f.
  
    unfold derives, projecting, field_pred; intros.
    rewrite H; auto.
   Qed.

  Lemma projecting_eq : forall x a b G t,
    [ G ] |-- projecting x a ->
    [ G ] |-- projecting x b ->
    [ const (a = b) :: G ] |-- t ->
    [ G ] |-- t.

    intros.
    cut_with (projecting x a); auto.
    cut_with (projecting x b); [ | apply derives_drop_head; auto ].
    clear H H0.
    unfold projecting, derives; simpl; intuition.
    apply H1.
    simpl; intuition.
    subst a b.
    red; auto.
  Qed.

End LogicTheory.
