(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.
Require Import Modal.
Require Import GoedelLoeb.
Require Import Fields.
Require Import Adjectives.


Module ProgramSafetyTheory (W':WORLDS) (WS':WORLD_STEP with Module W:=W')
  : PROGRAM_SAFETY_THEORY with Module W:=W' with Module WS:=WS'.
  Module W:=W'.
  Module WS:=WS'.
  Module AT := AdjectiveTheory W'.

  Export WS.
  Export AT.

  Open Local Scope pred_scope.


  Definition step_or_halt (p:prog_desc) (k:prog_cont) (w:world) :=
     halted p k w \/ exists k', exists w', step p k w k' w'.

  Inductive stepN (p:prog_desc) : nat -> prog_cont -> world -> prog_cont -> world -> Prop :=
   | stepN_O: forall w k, stepN p 0 k w k w
   | stepN_S: forall n k k' k'' w w' w'',
          step p k w k' w' -> stepN p n k' w' k'' w'' -> stepN p (S n) k w k'' w''
   .

  Definition safeN (p: prog_desc) (n: nat) (k:prog_cont) (w:world) := 
     forall n' k' w', n' <= n -> stepN p n' k w k' w' -> step_or_halt p k' w'.

  Definition safe (p:prog_desc) (k:prog_cont) : pred :=
      fun w => safeN p (world_ttl w) k w.

  Definition guards (p:prog_desc) (k:prog_cont) (pre:pred) :=
     pre >=> safe p k.

 Lemma step_tick_n_commute : forall p n k k' w w',
     (exists w1, step p k w k' w1 /\ tick_world_n n w1 = Some w') <->
     (exists w2, tick_world_n n w = Some w2 /\ step p k w2 k' w').

    intros p n; induction n.
    split; intros.

    destruct H; simpl in *; intuition.
    replace w' with x; [ | congruence ].
    exists w; intuition.
    destruct H; simpl in *; intuition.
    replace w with x; [ | congruence ].
    exists w'; intuition.
    split; intros.
    destruct H; simpl in*; intuition.
    case_eq (tick_world x); intros;
      rewrite H in H1; try discriminate.
    destruct (tick_succeeds w).
    change (WS.W.WT.world_ttl w > 0).
    rewrite (step_preserves_ttl p k k' w x); auto.
    change (world_ttl x > 0).
    rewrite (tick_ttl x w0); auto.
    omega.
    rewrite H2.
    rename x0 into w1.
    destruct (IHn k k' w1 w').
    apply H3; clear H3 H4.
    exists w0; intuition.
    destruct (step_tick_commute p k k' w w0).
    destruct H3; clear H4.
    exists x; intuition.
    destruct H3.
    change (tick_world w = Some x0) in H3.
    replace w1 with x0; [ auto | congruence ].
    
    destruct H; simpl in *; intuition.
    case_eq (tick_world w); intros;
      rewrite H in H0; try discriminate.
    destruct (untick_n n w').
    destruct (IHn k k' w0 w').
    destruct H4; clear H3.
    exists x; intuition.
    destruct H4.
    cut (x0 = x1); intros.
    subst x1.
    destruct (step_tick_commute p k k' w x0).
    destruct H6; clear H5.
    exists w0; intuition.
    destruct H6.
    exists x1; intuition.
    change (tick_world x1 = Some x0) in H6.
    rewrite H6.
    auto.
    apply tick_n_inj with n w'; auto.
   Qed.

  Lemma tick_step_or_halt : forall p n k w w',
    tick_world_n n w = Some w' ->
    step_or_halt p k w -> step_or_halt p k w'.

    intros p n; induction n; simpl; intros.
    replace w' with w; [ auto | congruence ].
    case_eq (tick_world w); intros;
      rewrite H1 in H; try discriminate.
    apply IHn with w0; auto.
    clear H IHn.
    destruct H0.
    left.
    destruct (halt_tick p k w w0); auto.
    right.
    destruct H as [k' [x ?]].
    destruct (tick_succeeds x).
    change (WS.W.WT.world_ttl x > 0).
    rewrite <- (step_preserves_ttl p k k' w x); auto.
    change (world_ttl w > 0).
    rewrite (tick_ttl w w0); auto; omega.
    destruct (step_tick_commute p k k' w x0).
    destruct H2; clear H3.
    exists x; intuition.   
    destruct H2.
    change (tick_world w = Some x1) in H2.
    replace w0 with x1; [ | congruence ].
    exists k'; exists x0; auto.
   Qed.   

  Lemma tick_step_or_halt2 : forall p n k w w',
    tick_world_n n w = Some w' ->
    step_or_halt p k w' -> step_or_halt p k w.

    intros p n; induction n; simpl; intros.
    replace w with w'; [ auto | congruence ].
    case_eq (tick_world w); intros;
      rewrite H1 in H; try discriminate.
    destruct (untick_n 1 w').
    destruct (tick_n_split (S n) n 1 w w').
    omega.
    simpl.
    rewrite H1; auto.
    destruct H3.
    apply (IHn k _ _ H3).
    simpl in H4.
    case_eq (tick_world x0); intros;
      rewrite H5 in H4; try discriminate.
    replace w1 with w' in *; [ | congruence ].
    destruct H0.
    left.
    destruct (halt_tick p k _ _ H5).
    apply H7.
    auto.
    right.
    destruct H0 as [k' [x1 ?]].
    destruct (step_tick_commute p k k' x0 x1).
    destruct H7.
    exists w'; intuition.
    exists k'; exists x2; intuition.
   Qed.

  Lemma safe_later : forall p k,
    safe p k |-- |> (safe p k).

    unfold derives; simpl; intuition.
    rewrite laterM_spec in *.
    unfold safe, safeN, later in *; intros.
    unfold later_rel; red; intros.
    destruct H as [n Hn].
    destruct (untick_n (S n) w'0).
    apply tick_step_or_halt with (S n) x; auto.
    apply H0 with n'.
    rewrite (tick_n_ttl (S n) w w'); auto.
    omega.
    revert H H3 Hn; clear.
    generalize (S n); clear n.
    revert k k' w w' w'0 x.
    induction n'; intros.
    inversion H3; subst; clear H3.
    replace x with w.
    apply stepN_O.
    apply tick_n_inj with n w'0; auto.
    inversion H3; subst; clear H3.
    destruct (untick_n n w'1).
    generalize (IHn' _ _ _ _ _ _ _ H H2 H0).
    intros.
    apply stepN_S with k'0 x0; auto.
    clear H H2 H3.
    destruct (step_tick_n_commute p n k k'0 w w'1).
    destruct H2; clear H.
    exists w'; intuition.
    destruct H2.
    replace x0 with x1; auto.
    apply tick_n_inj with n w'1; auto.
  Qed.

  Definition with_step_rel (p:prog_desc) (k k':prog_cont) (w w':world) :=
    step p k w k' w'.

  Definition with_step (p:prog_desc) (k k':prog_cont) := mkNegModality (with_step_rel p k k').
  Lemma with_step_spec : forall p k k', modal_op (with_step p k k') = appNegMode (with_step_rel p k k').
    intros; auto.
   Qed.
  Lemma with_step_rel_spec : forall p k k', modal_rel (with_step p k k') = with_step_rel p k k'.
    intros; auto.
   Qed.
  Opaque with_step.

  Lemma with_step_later_commute : forall p k k' t,
     with_step p k k' (|> t) |-- |> (with_step p k k' t).

    intros; rewrite with_step_spec; rewrite laterM_spec.
    unfold derives; simpl; intuition.
    unfold later, appNegMode, appPosMode in *.
    intros.
    destruct H0 as [w0 [Hw1 Hw2]].
    destruct H as [n Hn].
    destruct (tick_n_succeeds (S n) w0).
    change (WS.W.WT.world_ttl w0 >= S n).
    rewrite <- (step_preserves_ttl p k k' w w0); auto.
    change (world_ttl w >= S n).
    rewrite (tick_n_ttl (S n) w w'); auto.
    omega.
    exists x; intuition.
    destruct (step_tick_n_commute p (S n) k k' w x).
    destruct H0; clear H2.
    exists w0; intuition.
    destruct H0.
    replace w' with x0; auto.
    congruence.
    apply Hw2.
    red; exists n; auto.
   Qed.

  Lemma with_step_everywhere : forall p k k' t,
    with_step p k k' (!t) |-- !t.

    intros p k k' t.
    rewrite with_step_spec.
    rewrite everywhere_spec.
    unfold everywhere, appPosMode, appNegMode.
    unfold derives; simpl; intuition.
    red in H.
    destruct H0 as [w'' ?]; intuition.
    apply H3.
    red.
    change (WS.W.WT.world_ttl w'' = world_ttl w').
    rewrite <- (step_preserves_ttl p k k' w w''); auto.
   Qed.

  Lemma with_step_everywhere2 : forall p k k' t t',
    [ with_step p k k' t' :: !t :: nil ] |-- with_step p k k' (!t).

    intros p k k' t t'.
    rewrite with_step_spec.
    rewrite everywhere_spec.
    unfold everywhere, appPosMode, appNegMode.
    unfold derives; simpl; intuition.
    destruct H0 as [w' ?]; intuition.
    exists w'; intuition.
    apply H.
    red.
    rewrite <- H0.
    apply step_preserves_ttl with p k k'; auto.
   Qed.

  Lemma safe_step : forall p k k',
    with_step p k k' (|> safe p k') |-- safe p k.

    intros p k k'.
    rewrite with_step_spec; rewrite laterM_spec.
    unfold derives; simpl; intuition.
    unfold later, appNegMode, appPosMode, safe in *.
    destruct H0 as [w' ?]; intuition.
    unfold safeN in *; intros.
    inversion H3; subst; clear H3.
    right; exists k'; exists w'; auto.
    red in H0. 
    destruct (step_deterministic _ _ _ _ _ _ _ H0 H4).
    subst k'1 w'1.
    destruct (tick_succeeds w'0).
    replace (world_ttl w'0) with (world_ttl w); auto.
    omega.
    change (WS.W.WT.world_ttl w = world_ttl w'0).
    rewrite (step_preserves_ttl p k k' w w'); auto.
    revert H5; clear.
    revert k' k'0 w'0 w'; induction n; simpl; intros.
    inversion H5; subst; auto.
    inversion H5; subst; auto.
    rewrite (step_preserves_ttl p k' k'1 w' w'1); auto.
    refine (IHn _ _ _ _ H1).
    destruct (tick_succeeds w').
    change (WS.W.WT.world_ttl w' > 0).
    rewrite <- (step_preserves_ttl p k k' w w'); auto.
    change (world_ttl w > 0).
    omega.
    assert (stepN p n k' x0 k'0 x).
    revert H5 x H3 x0 H6; clear.
    intro H; induction H; simpl; intros.
    replace x with x0; try congruence.
    apply stepN_O.
    destruct (tick_succeeds w').    
    change (WS.W.WT.world_ttl w' > 0).
    rewrite <- (step_preserves_ttl p k k' w w'); auto.
    change (world_ttl w > 0).
    rewrite (tick_ttl w x0); auto.
    omega.
    destruct (step_tick_commute p k k' w x1).
    destruct H2; clear H4.
    exists w'; split; auto.
    destruct H2.
    change (tick_world w = Some x2) in H2.
    replace x2 with x0 in H4; [ | congruence ].
    eapply stepN_S; [ apply H4 | ].
    apply IHstepN; auto.
    assert (step_or_halt p k'0 x).
    apply H2 with x0 n; auto.
    exists 0; simpl; rewrite H6; auto.
    assert (S n <= S (world_ttl x0)).
    rewrite <- (tick_ttl w' x0); auto.
    change (world_ttl w') with (WS.W.WT.world_ttl w').
    rewrite <- (step_preserves_ttl p k k' w w'); auto.
    omega.
    destruct H8.
    left.
    destruct (halt_tick p k'0 w'0 x); auto.
    destruct H8 as [k'2 [w'2 ?]].
    right.
    exists k'2.
    destruct (step_tick_commute p k'0 k'2 w'0 w'2).
    destruct H10; clear H9.
    exists x; auto.
    destruct H10.
    exists x1; auto.
   Qed.

  Lemma step_guard : forall p k e G,
     ( forall t, In t G -> t |-- []!t ) ->
     [ e :: G ] |-- Ex_ e':pred, (Ex_ k':prog_cont, |> (guards p k' e') && with_step p k k' (|> e')) ->
     [ G ] |-- guards p k e.

    intros; unfold guards.
    cut ([ e :: G ] |-- safe p k).
    unfold sub.
    clear H0.
    revert e H.
    induction G; intros.
    repeat apply modal_N; auto.
    cut ([ G ] |-- []!(e && a --> safe p k)).
    intros.
    cut_with ([]!a && []!(e && a --> safe p k)).
    rewrite <- pmodal_conj; auto.
    rewrite <- pmodal_conj; auto.
    apply derives_weaken_head.
    apply derives_modal.
    apply derives_modal.
    intuition_d.
    apply_at 2.
    split_d; assumed.
    split_d.
    apply_d H; auto.
    clear_at 0.
    apply H1.
    apply IHG.
    intros.
    apply H.
    simpl; auto.
    elim_at 0.
    apply H0.

    refine (derives_cut _ _ _ _ H0).
    elim_at 0; intro e'.
    elim_at 0; intro k'.
    apply_d (safe_step p k k').
    apply derives_weaken_head.
    elim_at 0.
    cut_with (with_step p k k' (|> (guards p k' e' && e'))).
    apply derives_weaken_head.
    apply derives_modal.
    apply derives_modal.
    elim_at 0.
    unfold guards.
    apply sub_project_herenow.

    rewrite with_step_spec.
    unfold appNegMode.
    unfold derives; simpl; intuition.
    destruct H1 as [w' ?]; intuition.
    exists w'; intuition.

    rewrite pmodal_conj; auto.
    split; auto.

    unfold guards, sub in *.
    rewrite later_nec_later.
    rewrite <- every_later_commute.
    rewrite later_nec_later in H2.
    rewrite <- every_later_commute in H2.

    rewrite everywhere_spec in H2.
    rewrite everywhere_spec.
    unfold everywhere, appPosMode in H2 |- *.
    intros; apply H2.
    unfold everywhere_rel in *.
    rewrite <- H1.
    refine (step_preserves_ttl p k k' w w' _); auto.
   Qed.

End ProgramSafetyTheory.
