(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Add LoadPath "logic".

Require Import ModelSpec.
Require Import Worlds.
Require Import Logic.
Require Import Modal.

Module LogicTactics (W':WORLDS) 
   : LOGIC_TACTICS with Module W:=W'.

   Module W := W'.
   Module MT := ModalTheory W.

   Export MT.

   Open Local Scope pred_scope.

   Inductive ModalImpl : Type :=
   | ModalImpl_lift : pred -> ModalImpl
   | ModalImpl_modal_pos : modality pol_pos -> ModalImpl -> ModalImpl
   | ModalImpl_modal_neutral : modality pol_neutral -> ModalImpl -> ModalImpl
   | ModalImpl_implies : pred -> pred -> ModalImpl
   .

   Fixpoint interpModalImpl (x:ModalImpl) : pred :=
     match x with
       | ModalImpl_modal_pos m x' => modal_op m (interpModalImpl x')
       | ModalImpl_modal_neutral m x' => modal_op m (interpModalImpl x')
       | ModalImpl_lift p => p
       | ModalImpl_implies p q => implies p q
     end.

   Fixpoint apply_modals (l:list { z:polarity & { m:modality z | is_pos z }}) (p:pred) { struct l } : pred :=
     match l with
       | nil => p
       | (existT _ (exist m _)) :: xs => m (apply_modals xs p)
     end.

   Fixpoint do_pushdown_modal (x:ModalImpl) (l:list { z:polarity & { m:modality z | is_pos z }}) { struct x } : pred :=
       match x with
         | ModalImpl_lift z => apply_modals l z
         | ModalImpl_modal_pos m z => do_pushdown_modal z (l ++ (@existT _ _ pol_pos (@exist _ _ m I))::nil)
         | ModalImpl_modal_neutral m z => do_pushdown_modal z (l ++ (@existT _ _ pol_neutral (@exist _ _ m I))::nil)
         | ModalImpl_implies p q => implies (apply_modals l p) (apply_modals l q)
       end.

   Fixpoint split_positive_modal_head (x:ModalImpl) : (list { z:polarity & { m:modality z | is_pos z }} * ModalImpl) :=
     match x with
       | ModalImpl_modal_pos m x' =>
         let (l, z) := split_positive_modal_head x' in ((@existT _ _ pol_pos (@exist _ _ m I)) :: l, z)
       | ModalImpl_modal_neutral m x' =>
         let (l, z) := split_positive_modal_head x' in ((@existT _ _ pol_neutral (@exist _ _ m I)) :: l, z)
       | _ => (nil, x)
     end.

   Lemma split_positive_modal_head_sound : forall x z l,
     split_positive_modal_head x = (l, z) ->
     interpModalImpl x = apply_modals l (interpModalImpl z).

     induction x; simpl; intros.
     injection H; intros; subst; simpl; auto.
     case_eq (split_positive_modal_head x); intros.
     rewrite H0 in H.
     injection H; intros; subst; simpl.
     rewrite IHx with z l0; auto.
     case_eq (split_positive_modal_head x); intros.
     rewrite H0 in H.
     injection H; intros; subst; simpl.
     rewrite IHx with z l0; auto.
     injection H; intros; subst; simpl; auto.
   Qed.

   Lemma pullout_modal_aux1 : forall l p q,
     apply_modals l (p --> q) |--
     apply_modals l p --> apply_modals l q.

     intro l; induction l; simpl; intros.
     assumed.
     destruct a; destruct s.
     apply_d modal_K.
     apply derives_modal.
     apply IHl.
   Qed.

   Lemma pullout_modal_aux : forall l p q G,
     [ G ] |-- apply_modals l (p --> q) ->
     [ G ] |-- apply_modals l p --> apply_modals l q.

     intros; apply_d pullout_modal_aux1; auto.
   Qed.

   Lemma apply_modals_app : forall l1 l2 p,
     apply_modals (l1 ++ l2) p = apply_modals l1 (apply_modals l2 p).

     intro l1; induction l1; simpl; intros; auto.
     destruct a; destruct s.
     rewrite IHl1; auto.
   Qed.

   Lemma apply_modals_derives : forall l t1 t2,
     t1 |-- t2 ->
     apply_modals l t1 |-- apply_modals l t2.

     intro l; induction l; simpl; intros; auto.
     destruct a.
     destruct s.
     apply derives_modal.
     refine (IHl _ _ _); auto.
   Qed.

   Lemma apply_modals_all : forall l G t,
     derives G t ->
     derives (List.map (apply_modals l) G) (apply_modals l t).

     intros l G; revert l; induction G; simpl; intros.
     induction l; simpl; intros; auto.
     destruct a; destruct s.
     apply modal_N; auto.
     cut_with (apply_modals l (a --> t)).
     weaken_to (apply_modals l (a --> t) :: apply_modals l a :: nil).
     induction l; simpl; intros; auto.
     destruct a0; destruct s.
     cut_with (x0 (apply_modals l (a --> t) && apply_modals l a)).
     apply derives_weaken_head.
     apply derives_modal.
     apply conj_elim.
     apply IHl.
     rewrite pmodal_conj; auto.
     clear_at 0.
     apply IHG; auto.
   Qed.

   Lemma pushdown_aux : forall x l1 l2,
     [ apply_modals l1 (apply_modals l2 (interpModalImpl x)) :: nil ] |-- do_pushdown_modal x (l1 ++ l2).


     intro x; induction x; simpl; intros; auto.
     rewrite apply_modals_app; assumed.

     rewrite app_ass.
     apply_d (IHx l1).
     apply apply_modals_derives.
     rewrite apply_modals_app.
     apply apply_modals_derives.
     simpl; assumed.

     rewrite app_ass.
     apply_d (IHx l1).
     apply apply_modals_derives.
     rewrite apply_modals_app.
     apply apply_modals_derives.
     simpl; assumed.

     apply implies_intro.
     rewrite <- apply_modals_app.
     cut_with (apply_modals (l1++l2) p --> apply_modals (l1++l2) p0).
     apply implies_elim; assumed.
     apply_d pullout_modal_aux1.
     assumed.
   Qed.

   Lemma pushdown_modal_sound : forall x,
     interpModalImpl x |-- do_pushdown_modal x nil.

     intros x; exact (pushdown_aux x nil nil).
   Qed.


  Inductive LogicRep : Type :=
    | LR_Lift : pred -> LogicRep
    | LR_Bot : LogicRep
    | LR_Top : LogicRep
    | LR_Disj : LogicRep -> LogicRep -> LogicRep
    | LR_Conj : LogicRep -> LogicRep -> LogicRep
    | LR_Impl : LogicRep -> LogicRep -> LogicRep
    | LR_Ex   : forall A:Type, (A->pred) -> LogicRep
    | LR_All  : forall A:Type, (A->pred) -> LogicRep
    | LR_Neg  : LogicRep -> LogicRep
    | LR_Const : Prop -> LogicRep
    | LR_Proj  : forall x:field, field_type x -> LogicRep
    | LR_FPred : forall x:field, (field_type x -> pred) -> LogicRep
    | LR_PMod  : modality pol_pos -> LogicRep -> LogicRep
    | LR_NMod  : modality pol_neg -> LogicRep -> LogicRep
    | LR_BMod  : modality pol_neutral -> LogicRep -> LogicRep
    .

  Fixpoint interpLogicRep (x:LogicRep) : pred :=
    match x with
    | LR_Lift p         => p
    | LR_Bot            => bot
    | LR_Top            => top
    | LR_Disj p q       => disj (interpLogicRep p) (interpLogicRep q)
    | LR_Conj p q       => conj (interpLogicRep p) (interpLogicRep q)
    | LR_Impl p q       => implies (interpLogicRep p) (interpLogicRep q)
    | LR_Ex A f         => ex f
    | LR_All A f        => all f
    | LR_Neg p          => neg (interpLogicRep p)
    | LR_Const P        => const P
    | LR_Proj x a       => projecting x a
    | LR_FPred x f      => field_pred x f
    | LR_PMod m p       => m (interpLogicRep p)
    | LR_NMod m p       => m (interpLogicRep p)
    | LR_BMod m p       => m (interpLogicRep p)
    end.

  Inductive CxtRep : Type :=
    | CR_Goal  : pred -> CxtRep
    | CR_Hyp   : pred -> CxtRep -> CxtRep
    | CR_Const : Prop -> CxtRep -> CxtRep
    | CR_Ex    : forall A:Type, (A->pred) -> CxtRep -> CxtRep
    .

  Fixpoint interp_cxt (Cxt:CxtRep) (l:list pred) { struct Cxt } : Prop :=
    match Cxt with
    | CR_Goal g     => derives l g
    | CR_Hyp H C'   => interp_cxt C' (l ++ H::nil)
    | CR_Const P C' => P -> interp_cxt C' l
    | CR_Ex A f C'  => forall x:A, interp_cxt C' (l ++ (f x::nil))
    end.

  Definition interp_simpl_cxts (Cxts:list CxtRep) (l:list pred) (P:Prop) :=
    fold_right (fun Cxt X => interp_cxt Cxt l /\ X) P Cxts.

  Lemma interp_simpl_cxts_in : forall Cxts l P c,
     In c Cxts ->
     interp_simpl_cxts Cxts l P ->
     interp_cxt c l.

    intro Cxts; induction Cxts; simpl; intros.
    contradiction.
    destruct H; subst; intuition.
    apply IHCxts with P; auto.
   Qed.

  Lemma interp_simpl_cxts_in2 : forall Cxts l (P:Prop),
     P ->
     (forall c, In c Cxts -> interp_cxt c l) ->
     interp_simpl_cxts Cxts l P.

     intro Cxts; induction Cxts; simpl; intros; auto.
    Qed.

  Lemma interp_simpl_cxts_tail : forall Cxts l P,
     interp_simpl_cxts Cxts l P -> P.

     intro Cxts; induction Cxts; simpl; intros; intuition.
     apply IHCxts with l; auto.
    Qed.

  Lemma interp_simpl_cxts_weaken_tail : forall Cxts l (P P':Prop),
       (P -> P') ->
       interp_simpl_cxts Cxts l P -> interp_simpl_cxts Cxts l P'.

     intros Cxts; induction Cxts; simpl; intros; intuition.
     apply IHCxts with P; auto.
    Qed.

  Lemma interp_simpl_cxts_weaken : forall C' C l P,
     (forall c, In c C' -> In c C) ->
     interp_simpl_cxts C l P ->
     interp_simpl_cxts C' l P.

     intro C'; induction C'; simpl; intros.
     apply (interp_simpl_cxts_tail _ _ _ H0).
     split.
     assert (In a C).
     apply H; auto.
     revert H0 H1; clear.
     induction C; simpl; intros; [ contradiction | ].
     destruct H1; subst; intuition.
     apply IHC' with C; auto.
    Qed.

  Lemma interp_simpl_cxts_app : forall C1 C2 l P,
     interp_simpl_cxts (C1++C2) l P = 
     interp_simpl_cxts C1 l (interp_simpl_cxts C2 l P).

     unfold interp_simpl_cxts; intros.
     rewrite fold_right_app.
     auto.
    Qed.

  Fixpoint do_simpl_hyp1 (H:LogicRep) (Cxts:list CxtRep) { struct H } : list CxtRep :=
     match H with
     | LR_Bot       => nil
     | LR_Top       => Cxts
     | LR_Disj p q  => do_simpl_hyp1 p Cxts ++ do_simpl_hyp1 q Cxts
     | LR_Conj p q  => do_simpl_hyp1 p (do_simpl_hyp1 q Cxts)
     | LR_Ex A ex_f => List.map (CR_Ex A ex_f) Cxts
     | LR_Const P   => List.map (CR_Const P) Cxts
     | _            => List.map (CR_Hyp (interpLogicRep H)) Cxts
     end.

  Lemma do_simpl_hyp1_in_thru : forall x l1 l2,
     (forall a, In a l1 -> In a l2) ->
     (forall a, In a (do_simpl_hyp1 x l1) -> In a (do_simpl_hyp1 x l2)).

    intro x; induction x; simpl; intros; auto; try (
       match goal with [ |- In _ (List.map ?C _) ] =>
          destruct (in_map_iff C l2 a); apply H2;
          destruct (in_map_iff C l1 a);
          let x := fresh in
          destruct (H3 H0) as [x ?];
          exists x; auto; intuition
        end).

    apply in_or_app.
    destruct (in_app_or _ _ _ H0).
    left; eapply IHx1; [ apply H | apply H1 ].
    right; eapply IHx2; [ apply H | apply H1 ].
    eapply IHx1; [ | apply H0 ].
    intros.
    eapply IHx2; [ | apply H1 ].
    auto.
   Qed.

  Lemma do_simpl_hyp1_in1 : forall x l1 l2 a,
     In a (do_simpl_hyp1 x (l1 ++ l2)) ->
     In a (do_simpl_hyp1 x l1 ++ do_simpl_hyp1 x l2).

     intro x; induction x; simpl; intros; auto;
        try (rewrite map_app in H; auto; fail).
     destruct (in_app_or _ _ _ H).
     destruct (in_app_or _ _ _ (IHx1 _ _ _ H0)).
     apply in_or_app; left; apply in_or_app; left; auto.
     apply in_or_app; right; apply in_or_app; left; auto.
     destruct (in_app_or _ _ _ (IHx2 _ _ _ H0)).
     apply in_or_app; left; apply in_or_app; right; auto.
     apply in_or_app; right; apply in_or_app; right; auto.
     apply in_or_app.
     cut (In a (do_simpl_hyp1 x1 (do_simpl_hyp1 x2 l1 ++ do_simpl_hyp1 x2 l2))).
     intro H1.
     destruct (in_app_or _ _ _ (IHx1 _ _ _ H1)); auto.
     revert H; apply do_simpl_hyp1_in_thru.
     intros.
     apply in_or_app.
     destruct (in_app_or _ _ _ (IHx2 _ _ _ H)); auto.
    Qed.

  Lemma do_simpl_hyp1_in2 : forall x l1 l2 a,
     In a (do_simpl_hyp1 x l1 ++ do_simpl_hyp1 x l2) ->
     In a (do_simpl_hyp1 x (l1 ++ l2)).

     intro x; induction x; simpl; intros; auto; 
       try (rewrite map_app; auto; fail).
     apply in_or_app.
     destruct (in_app_or _ _ _ H).
     destruct (in_app_or _ _ _ H0).
     left; apply IHx1; apply in_or_app; auto.
     right; apply IHx2; apply in_or_app; auto.
     destruct (in_app_or _ _ _ H0).
     left; apply IHx1; apply in_or_app; auto.
     right; apply IHx2; apply in_or_app; auto.
     destruct (in_app_or _ _ _ H); revert H0;
       apply do_simpl_hyp1_in_thru; intros;
       apply IHx2; apply in_or_app; auto.
    Qed.

  Lemma derives_app_unit : forall l p q,
     [ p :: l ] |-- q ->
     [ l ++ p :: nil ] |-- q.

     intros l p q; apply derives_weaken.
     simpl; intros; apply in_or_app; simpl; intuition.
    Qed.

  Lemma derives_app_unit2 : forall l p q,
     [ l ++ p :: nil ] |-- q ->
     [ p :: l ] |-- q.

     intros l p q; apply derives_weaken.
     simpl; intros.
     destruct (in_app_or _ _ _ H); simpl in *; intuition.
    Qed.

  Lemma interp_cxt_weaken : forall c l1 l2,
     (forall x, In x l1 -> In x l2) ->
     interp_cxt c l1 -> interp_cxt c l2.

     intro c; induction c; simpl; intros.
     apply derives_weaken with l1; auto.
     apply IHc with (l1++p::nil); simpl; auto; intros.
     apply in_or_app.
     destruct (in_app_or _ _ _ H1); intuition.
     apply IHc with l1; simpl; auto.
     apply IHc with (l1 ++ p x::nil); simpl; auto.
     simpl; intros.
     apply in_or_app.
     destruct (in_app_or _ _ _ H1); intuition.
    Qed.

  Lemma interp_cxt_app_unit : forall c p l,
     interp_cxt c (p :: l) ->
     interp_cxt c (l ++ (p::nil)).

     intros c p l; apply interp_cxt_weaken; intros.
     apply in_or_app; simpl in *; intuition.
    Qed.

  Lemma interp_cxt_app_unit2 : forall c p l,
     interp_cxt c (l ++ (p::nil)) ->
     interp_cxt c (p :: l).

     intros c p l; apply interp_cxt_weaken; intros.
     simpl; destruct (in_app_or _ _ _ H); simpl in *; intuition.
    Qed.

  Lemma interp_cxt_drop_right : forall c l p,
     interp_cxt c l -> interp_cxt c (l ++ p::nil).
     
     intros c l p; apply interp_cxt_weaken.
     simpl; intros.
     apply in_or_app; auto.
    Qed.

  Lemma interp_cxt_bot : forall c G,
     interp_cxt c (bot :: G).

    intro c; induction c; simpl; intros; try (apply IHc; auto).
    apply bot_elim.
   Qed.

  Lemma interp_cxt_disj : forall c p q G,
     interp_cxt c (p :: G) ->
     interp_cxt c (q :: G) ->
     interp_cxt c (p || q :: G).

    intro c; induction c; simpl; intros; try (apply IHc; auto).
    apply disj_elim; auto.
   Qed.

  Lemma interp_cxt_conj : forall c p q G,
     interp_cxt c (p :: q :: G) ->
     interp_cxt c (p && q :: G).

     intro c; induction c; simpl; intros; try (apply IHc; auto).
     apply conj_elim; auto.
    Qed.

  Lemma interp_cxt_exist : forall c A (f:A->pred) G,
     (forall x:A, interp_cxt c (f x :: G)) ->
     interp_cxt c (ex f :: G).

     intro c; induction c; simpl; intros; try (apply IHc; auto).
     apply exist_elim; intuition.
    Qed.

  Lemma interp_cxt_const : forall c (P:Prop) G,
     (P -> interp_cxt c G) ->
     interp_cxt c (const P :: G).

     intro c; induction c; simpl; intros; try (apply IHc; auto).
     apply const_elim; intuition.
    Qed.

  Lemma interp_simpl_cxts_hyp : forall Cxts p l P,
      interp_simpl_cxts (List.map (CR_Hyp p) Cxts) l P ->
      interp_simpl_cxts Cxts (l ++ p::nil) P.

     intro Cxts; induction Cxts; simpl; intuition.
    Qed.

  Lemma interp_simpl_cxts_hyp1_in : forall H Cxts a l P',
     In a Cxts ->
     interp_simpl_cxts (do_simpl_hyp1 H Cxts) l P' ->
     interp_cxt a (l ++ interpLogicRep H :: nil).

     intro H; induction H; simpl; intuition; try 
       (apply interp_simpl_cxts_in with Cxts P'; [ | apply interp_simpl_cxts_hyp ]; auto).
     
     clear H; induction Cxts; simpl in *; intuition; subst; auto.
     apply interp_cxt_app_unit; apply interp_cxt_bot.
     clear H; induction Cxts; simpl in *; intuition; subst; auto.
     apply interp_cxt_drop_right; auto.

     clear a H1.
     revert l P' H2; induction Cxts; simpl in *; intuition; subst; auto.
     eapply interp_simpl_cxts_tail; apply H2.
     apply interp_cxt_app_unit.
     apply interp_cxt_disj; apply interp_cxt_app_unit2.
     red in H2.
     rewrite fold_right_app in H2.
     refine (IHLogicRep1 (a::Cxts) a l _ _ H2); simpl; auto.
     red in H2.
     rewrite fold_right_app in H2.
     generalize (interp_simpl_cxts_tail _ _ _ H2); intro H3.
     refine (IHLogicRep2 (a::Cxts) a l _ _ H3); simpl; auto.
     eapply IHCxts.
     revert H2; apply interp_simpl_cxts_weaken.
     intros.
     apply in_or_app.
     destruct (in_app_or _ _ _ H1).
     left.
     apply (do_simpl_hyp1_in2 H (a::nil) Cxts c).
     apply in_or_app; auto.
     right.
     apply (do_simpl_hyp1_in2 H0 (a::nil) Cxts c).
     apply in_or_app; auto.

     clear a H1.
     revert l P' H2; induction Cxts; simpl in *; intuition; subst; auto.
     eapply interp_simpl_cxts_tail; apply H2.
     apply interp_cxt_app_unit.
     apply interp_cxt_conj.
     apply interp_cxt_weaken with (interpLogicRep H :: l ++ (interpLogicRep H0 :: nil)).
     simpl; intuition.
     destruct (in_app_or _ _ _ H3); simpl in *; intuition.
     apply (IHLogicRep2 (a::Cxts) a (interpLogicRep H :: l) P'); simpl; auto.
     apply interp_simpl_cxts_in2.
     refine (interp_simpl_cxts_tail _ _ _ H2).
     simpl; intros.
     apply interp_cxt_app_unit2.
     apply IHLogicRep1 with (do_simpl_hyp1 H0 (a::Cxts)) P'; auto.

     apply IHCxts.
     revert H2; apply interp_simpl_cxts_weaken; simpl; intro c.
     apply do_simpl_hyp1_in_thru; intro c'.
     apply do_simpl_hyp1_in_thru; simpl; intuition.

     clear a H.
     revert l P' H0; induction Cxts; simpl in *; intuition; subst; auto.
     apply interp_cxt_app_unit.
     apply interp_cxt_exist.
     intros.
     apply interp_cxt_app_unit2.
     apply H.

     clear a H.    
     revert l P' H0; induction Cxts; simpl in *; intuition; subst; auto.
     apply interp_cxt_app_unit.
     apply interp_cxt_const.
     auto.
    Qed.

  Lemma interp_simpl_cxts_hyp1_head : forall H a l P,
     interp_simpl_cxts (do_simpl_hyp1 H (a::nil)) l P ->
     interp_cxt a (l ++ interpLogicRep H :: nil).

     intros.
     apply interp_simpl_cxts_hyp1_in with (a::nil) P; simpl; auto.
    Qed.

  Lemma do_simpl_hyp1_sound : forall Cxts H l P,
     interp_simpl_cxts (do_simpl_hyp1 H Cxts) l P ->
     interp_simpl_cxts Cxts (l ++ (interpLogicRep H :: nil)) P.

     intro Cxts; induction Cxts; simpl; intros.
     apply (interp_simpl_cxts_tail _ _ _ H0).
     assert (interp_simpl_cxts (do_simpl_hyp1 H (a::nil) ++ do_simpl_hyp1 H Cxts) l P).
     revert H0; apply interp_simpl_cxts_weaken.
     apply (do_simpl_hyp1_in2 H (a::nil) Cxts).
     rewrite interp_simpl_cxts_app in H1.
     split.
     eapply interp_simpl_cxts_hyp1_head.
     apply H1.
     apply IHCxts.
     apply (interp_simpl_cxts_tail _ _ _ H1).
    Qed.

  Fixpoint do_simpl_hyps (Hyps:list LogicRep) (Cxts:list CxtRep) { struct Hyps } : list CxtRep :=
    match Hyps with
    | H::Hyps' => do_simpl_hyps Hyps' (do_simpl_hyp1 H Cxts)
    | nil => Cxts
    end.

  Lemma do_simpl_hyps_sound : forall Hyps Cxts l P,
      interp_simpl_cxts (do_simpl_hyps (rev Hyps) Cxts) l P ->
      interp_simpl_cxts Cxts (l ++ List.map interpLogicRep Hyps) P.

     intro Hyps; pattern Hyps; apply rev_ind; clear Hyps; simpl; intros; auto.
     rewrite <- app_nil_end; auto.
     rewrite map_app; simpl.
     rewrite <- app_ass.
     rewrite rev_unit in H0.
     simpl in H0.
     apply do_simpl_hyp1_sound.
     apply H; auto.
    Qed.

  Definition simpl_hyps (Hyps:list LogicRep) (t:pred) : Prop :=
    interp_simpl_cxts (do_simpl_hyps (rev Hyps) ((CR_Goal t) :: nil)) nil True.

  Lemma simpl_hyps_sound : forall Hyps t,
    simpl_hyps Hyps t -> derives (List.map interpLogicRep Hyps) t.

    unfold simpl_hyps; intros.
    destruct (do_simpl_hyps_sound Hyps ((CR_Goal t)::nil) nil True); intuition.
   Qed.

  Fixpoint simpl_goal_rep (Hyps:list LogicRep) (g:LogicRep) { struct g } : Prop :=
    match g with
    | LR_Top      => True
    | LR_Conj p q => simpl_goal_rep Hyps p /\ simpl_goal_rep Hyps q
    | LR_Impl p q => simpl_goal_rep (p :: Hyps) q
    | LR_All A f  => forall x:A, simpl_hyps Hyps (f x)
    | _ => simpl_hyps Hyps (interpLogicRep g)
    end.

  Theorem simpl_goal_rep_sound : forall Hyps g,
    simpl_goal_rep Hyps g -> derives (List.map interpLogicRep Hyps) (interpLogicRep g).

    intros Hyps g; revert Hyps; induction g; intros; simpl;
       try refine (simpl_hyps_sound _ _ H).
    apply top_intro.
    simpl in H; destruct H.
    apply conj_intro; auto.
    apply implies_intro.
    simpl in H.
    refine (IHg2 (g1::Hyps) _); auto.
    simpl in H.
    apply forall_intro.
    intros; refine (simpl_hyps_sound _ _ (H x)).
   Qed.

End LogicTactics.
