(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Require Import Bool.

Require Export Maps.
Require Export Machine.
Require Export simple_tc.

(* This module signature collects together all the proof
   obligations that are required to prove that the above
   typechecking algorithm is sound.
 *)

Module Type SIMPLE_TC_SAFETY.
  Import SimpleTC.

  Parameter ty : Type.

  Parameter t_nil      : ty.
  Parameter t_list     : ty -> ty.
  Parameter t_listcons : ty -> ty.

  Definition env := map ty.
  Definition program_typing := map env.

  (* The subtyping relation for types,
     together with the properties required
     to prove soundness of the checking
     algorithm.
   *)
  Parameter subtype : ty -> ty -> Prop.
  
  Axiom sub_refl : forall t, subtype t t.

  Axiom sub_trans : forall t1 t2 t3,
     subtype t1 t2 ->
     subtype t2 t3 ->
     subtype t1 t3.

  Axiom sub_nil  : forall t,
     subtype t_nil (t_list t).

  Axiom sub_list : forall t1 t2,
     subtype t1 t2 ->
     subtype (t_list t1) (t_list t2).

  Axiom sub_listmixed : forall t1 t2,
     subtype t1 t2 ->
     subtype (t_listcons t1) (t_list t2).

  Axiom sub_listcons : forall t1 t2,
     subtype t1 t2 ->
     subtype (t_listcons t1) (t_listcons t2).
  
  Hint Resolve sub_refl sub_trans sub_nil
     sub_list sub_listmixed sub_listcons
     : typing.

  (* Subtyping for environments is defined in
     a straightforward way from regular subtyping.
   *)
  Definition env_sub (e e':env) :=
    forall v t', e'#v = Some t' ->
       exists t, e#v = Some t /\ subtype t t'.

  (* The Hoare "triple" relation specialized to
     the types for this typechecker.  
   *)
  Parameter check_instr :
    program -> program_typing -> env -> instr -> env -> Prop.

  Axiom check_instr_seq : forall p pt e e' e'' i1 i2,
      check_instr p pt e  i1 e' ->
      check_instr p pt e' i2 e'' ->
      check_instr p pt e  (i1;;i2) e''.

  Axiom check_instr_branch_list : forall p pt e e1 v l t,
      e#v = Some (t_list t) ->
      pt#l = Some e1 ->
      env_sub (e#v <- t_nil) e1 ->
      check_instr p pt
        e (instr_branch_if_nil v l) (e#v <- (t_listcons t)).

  Axiom check_instr_branch_listcons : forall p pt e l v t,
      e#v = Some (t_listcons t) ->
      check_instr p pt
         e (instr_branch_if_nil v l) e.

  Axiom check_instr_branch_nil : forall p pt e v l e1 e2,
      e#v = Some t_nil ->
      pt#l = Some e1 ->
      env_sub e e1 ->
      check_instr p pt
         e (instr_branch_if_nil v l) e2.

  Axiom check_instr_fetch_0: forall p pt e v1 v2 t,
      e#v1 = Some (t_listcons t) ->
      check_instr p pt
         e (instr_fetch_field v1 0 v2) (e#v2 <- t).

  Axiom check_instr_fetch_1: forall p pt e v1 v2 t,
      e#v1 = Some (t_listcons t) ->
      check_instr p pt
         e (instr_fetch_field v1 1 v2) (e#v2 <- (t_list t)).

  Axiom check_instr_cons: forall p pt e v1 v2 v3 t1 t2 t3,
      e#v1 = Some t1 ->
      e#v2 = Some t2 ->
      subtype t1 t3 ->
      subtype t2 (t_list t3) ->
      check_instr p pt
         e (instr_cons v1 v2 v3) (e#v3 <- (t_listcons t3)).

  (* A relation expressing sufficent preconditions for
     program safety of instruction blocks.
   *)
  Parameter check_block :
    program -> program_typing -> env -> instr -> Prop.

  Axiom check_block_halt: forall p pt e,
      check_block p pt e instr_halt.

  Axiom check_block_seq: forall p pt e i1 i2 e',
      check_instr p pt e i1 e' ->
      check_block p pt e' i2 ->
      check_block p pt e (i1 ;; i2).

  Axiom check_block_jump: forall p pt e v l e1,
      pt#l = Some e1 ->
      env_sub (unset _ e v) e1 ->
      check_block p pt e (instr_getlabel l v ;; instr_jump v).

  (* This relation expresses the fact that the program typing
     accuratly describes the given program.  That is, the sufficent
     conditions given in the program typing actually reflect the
     code found in the program.
   *)
  Definition check_blocks (p:program) (pt:program_typing) :=
   forall l i,
     p#l = Some i ->
     exists e,
       pt#l = Some e /\
       check_block p pt e i.

  (* This relation states that every label found in the program typing
     maps to some actual label in the program.
   *)

  Definition typing_dom_match (p: program) (pt: program_typing) : Prop :=
    forall l e, pt#l = Some e -> exists i, p#l = Some i.

  (* This is the main safety lemma.  It says that if all blocks
     check out with relation to the program typing, and the program
     typing domain checks out, and the entry label has no precondition,
     then the program is safe to run by jumping to the entry label.
   *)
  Axiom safety : forall (p:program) (pt: program_typing),
     check_blocks p pt ->
     typing_dom_match p pt ->
     pt#entry_label = Some (empty ty) ->
     safe_prog p.

End SIMPLE_TC_SAFETY.



(* This module signature expresses the goal theorem.
   It states that if a program passes the typechecker,
   then it is safe to run.  
 *)
Module Type SIMPLE_TC_SAFETY_THEOREM.
  Import SimpleTC.

  Axiom typecheck_safeprog : forall
     (p:program) (pt:program_typing),
     
     typecheck p pt = true -> safe_prog p.

End SIMPLE_TC_SAFETY_THEOREM.


(* This module functor proves the main safety theorem when
   given an implemenation of SIMPLE_TC_SAFETY.
   The proofs in here are basically just a straightforward
   unrolling of the typechecking algorithm to show that
   it correctly applies the rules.  The proofs from
   SIMPLE_TC_SAFETY do the work of showing that these
   rules are sufficent for safety.
 *)

Module SimpleTCSafetyTheorem (X:SIMPLE_TC_SAFETY)
  : SIMPLE_TC_SAFETY_THEOREM.

  Import SimpleTC.

  Fixpoint interp_ty (t:typ) : X.ty :=
    match t with
      | ty_nil => X.t_nil
      | ty_list t => X.t_list (interp_ty t)
      | ty_listcons t => X.t_listcons (interp_ty t)
    end.

  Definition interp_env (e:env) : X.env := map_fmap typ X.ty e interp_ty.
  Definition interp_pt (pt:program_typing) : X.program_typing
      := map_fmap env X.env pt interp_env.

  Lemma check_sub_sound : forall t1 t2,
     check_subtype t1 t2 = true ->
     X.subtype (interp_ty t1) (interp_ty t2).

    intros t1; induction t1; intros; destruct t2; simpl in *;
      try discriminate; auto; try apply X.sub_refl.
    apply X.sub_nil.
    apply X.sub_list; auto.
    apply X.sub_listmixed; auto.
    apply X.sub_listcons; auto.
   Qed.

  Lemma check_env_sub_sound : forall e1 e2,
    check_env_sub e1 e2 = true ->
    X.env_sub (interp_env e1) (interp_env e2).

    unfold check_env_sub; red; intros.
    destruct (fmap_eqn2 _ _ e2 interp_ty v t') as [t ?]; intuition.
    generalize (map_forall2_correct _ _ check_env_binding_sub e2 e1 v t H H2).
    intro H1.
    case_eq (e1#v); intros; rewrite H4 in H1;
      simpl in H1; try discriminate.
    exists (interp_ty t0); intuition.
    unfold interp_env; apply fmap_eqn; auto.
    rewrite <- H3.
    apply check_sub_sound; auto.
   Qed.

  Lemma lub_sound1 : forall t1 t2, X.subtype (interp_ty t1) (interp_ty (lub t1 t2)).
     intro t1; induction t1; intros; destruct t2; simpl; auto with typing.
   Qed.

  Lemma lub_sound2 : forall t2 t1, X.subtype (interp_ty t2) (interp_ty (lub t1 t2)).
    intros t2; induction t2; intros; destruct t1; simpl; auto with typing.
   Qed.

  Lemma check_instr_sound : forall p pt i e e',
     typecheck_instr pt e i = Some e' ->
     X.check_instr p (interp_pt pt) (interp_env e) i (interp_env e').

     intros p pt i; induction i; simpl; intros; try discriminate.
     case_eq (e#v); intros; rewrite H0 in H; try discriminate.
     destruct t.
     unfold typecheck_branch in H.
     case_eq (pt#l); intros; rewrite H1 in H; try discriminate.
     case_eq (check_env_sub e e0); intros; rewrite H2 in H; try discriminate.
     apply X.check_instr_branch_nil with (interp_env e0); auto.
     unfold interp_env; refine (fmap_eqn _ _ _ _ _ _ H0).
     unfold interp_pt; apply fmap_eqn; auto.
     apply check_env_sub_sound; auto.
     unfold typecheck_branch in H.
     case_eq (pt#l); intros; rewrite H1 in H; try discriminate.
     case_eq (check_env_sub (e#v <- ty_nil) e0); intros; rewrite H2 in H; try discriminate.
     simplify_eq H; intros; subst e'.
     unfold interp_env.
     rewrite fmap_update.
     simpl.
     apply X.check_instr_branch_list with (interp_env e0); auto.
     unfold interp_env; refine (fmap_eqn _ _ _ _ _ _ H0).
     unfold interp_pt; apply fmap_eqn; auto.     
     change X.t_nil with (interp_ty ty_nil).
     rewrite <- fmap_update.
     refine (check_env_sub_sound _ _ H2).
     replace e' with e; [ | congruence ].
     apply X.check_instr_branch_listcons with (interp_ty t); auto.
     refine (fmap_eqn _ _ _ _ _ _ H0).

     destruct n.
     case_eq (e#v); intros; rewrite H0 in H; try discriminate.
     destruct t; try discriminate.
     simplify_eq H; intros; subst e'.
     unfold interp_env.
     rewrite fmap_update.
     apply X.check_instr_fetch_0; auto.
     refine (fmap_eqn _ _ _ _ _ _ H0).
     destruct n; try discriminate.
     case_eq (e#v); intros; rewrite H0 in H; try discriminate.
     destruct t; try discriminate.
     simplify_eq H; intros; subst e'.
     unfold interp_env.
     rewrite fmap_update.
     simpl.
     apply X.check_instr_fetch_1; auto.
     refine (fmap_eqn _ _ _ _ _ _ H0).
     case_eq (e#v); intros; rewrite H0 in H; try discriminate.   
     case_eq (e#v0); intros; rewrite H1 in H; try discriminate.
     destruct t0.
     simplify_eq H; intros; subst e'.
     unfold interp_env; rewrite fmap_update.
     simpl.
     apply X.check_instr_cons with (interp_ty t) X.t_nil; auto with typing.
     refine (fmap_eqn _ _ _ _ _ _ H0). 
     refine (fmap_eqn _ _ _ _ _ _ H1).
     simplify_eq H; intros; subst e'.
     unfold interp_env; rewrite fmap_update.
     simpl.
     apply X.check_instr_cons with (interp_ty t) (X.t_list (interp_ty t0)); auto with typing.
     refine (fmap_eqn _ _ _ _ _ _ H0). 
     refine (fmap_eqn _ _ _ _ _ _ H1).
     apply lub_sound1; auto with typing.
     apply X.sub_list; apply lub_sound2.
     simplify_eq H; intros; subst e'.
     unfold interp_env; rewrite fmap_update.
     simpl.
     apply X.check_instr_cons with (interp_ty t) (X.t_listcons (interp_ty t0)); auto with typing.
     refine (fmap_eqn _ _ _ _ _ _ H0). 
     refine (fmap_eqn _ _ _ _ _ _ H1).
     apply lub_sound1; auto with typing.
     apply X.sub_listmixed; apply lub_sound2.
   
     case_eq (typecheck_instr pt e i1); intros; rewrite H0 in H; try discriminate.
     apply X.check_instr_seq with (interp_env e0).
     apply IHi1; auto.
     apply IHi2; auto.
   Qed.

  Lemma check_block_sound : forall p pt i e,
     typecheck_block pt e i = true ->
     X.check_block p (interp_pt pt) (interp_env e) i.

     intros p pt i; induction i; simpl; intros; try discriminate.
     apply X.check_block_halt.
     destruct i1; try discriminate;
       [
         case_eq (typecheck_instr pt e (instr_getlabel l v)); intros;
         rewrite H0 in H; simpl in H0; try discriminate;
         destruct i2; try discriminate;
         destruct (andb_prop _ _ H);
         destruct (var_eq v v0); try discriminate;
         subst v0; simpl in *;
         unfold typecheck_branch in H;
         case_eq (pt#l); intros; rewrite H3 in H; try discriminate;
         (apply X.check_block_jump with (interp_env e0); auto;
          [ refine (fmap_eqn _ _ _ _ _ _ H3)
          | unfold interp_env;
            rewrite <- fmap_unset;
            refine (check_env_sub_sound _ _ H)
          ])

       | | | | ];

       match goal with
         [ |- X.check_block _ _ _ (?I ;; i2) ] =>

         case_eq (typecheck_instr pt e I); intros;
         rewrite H0 in H; try discriminate;
         eapply X.check_block_seq;
           [ apply check_instr_sound; apply H0
           | apply IHi2; auto
           ]
       end.
   Qed.

  Lemma typecheck_blocks_sound : forall pt p,

     typecheck_blocks p pt = true ->
     X.check_blocks p (interp_pt pt).

    unfold typecheck_blocks; intros.
    unfold X.check_blocks; intros.
    generalize (map_forall2_correct _ _ _ _ _ l i H H0).
    unfold check_one_block.
    case_eq (pt#l); intros; try discriminate.
    exists (interp_env e); intuition.
    refine (fmap_eqn _ _ _ _ _ _ H1).
    apply check_block_sound; auto.
   Qed.  

  Lemma check_pt_domain_sound : forall pt p,
    check_pt_domain p pt = true ->
    X.typing_dom_match p (interp_pt pt).

    unfold check_pt_domain; red; intros.
    destruct (fmap_eqn2 _ _ _ _ _ _ H0) as [e0 ?]; intuition.
    generalize (map_forall2_correct _ _ _ _ _ l e0 H H2).
    unfold check_one_domain; intros.
    case_eq (p#l); intros.
    exists i; auto.
    rewrite H4 in H1.
    discriminate.
   Qed.    

  Theorem typecheck_safeprog : forall p pt,
    typecheck p pt = true -> safe_prog p.

    unfold typecheck; intros.
    destruct (andb_prop _ _ H) as [H1 H2]; clear H.
    destruct (andb_prop _ _ H1) as [H3 H4]; clear H1.
    apply X.safety with (interp_pt pt).
    apply typecheck_blocks_sound; auto.
    apply check_pt_domain_sound; auto.
    case_eq (pt#entry_label); intros; rewrite H in H2; try discriminate.
    destruct e; simpl in H2; try discriminate.
    refine (fmap_eqn _ _ _ _ _ _ H).
   Qed.

End SimpleTCSafetyTheorem.
