(* Copyright 2006, 2008 Andrew W. Appel, Robert Dockins and Xavier Leroy
 * This file is part of the List-machine Benchmark.
 * The List-machine Benchmark is licensed under GPLv3; see LICENSE.
 *)

Require Import Bool.
Require Import CompleteInduction.

Require Export simple_tc_safety.

Module SimpleTCSyntactic : SIMPLE_TC_SAFETY.

  Inductive typ : Set :=
  | ty_nil      : typ
  | ty_list     : typ -> typ
  | ty_listcons : typ -> typ
  .

  Definition ty := typ.
  Definition t_nil := ty_nil.
  Definition t_list := ty_list.
  Definition t_listcons := ty_listcons.

  Definition env := map ty.
  Definition program_typing := map env.

  Inductive subtype' : typ -> typ -> Prop :=
  | sub_refl' : forall t,
      subtype' t t
    
  | sub_nil'  : forall t,
      subtype' ty_nil (ty_list t)

  | sub_list' : forall t1 t2,
      subtype' t1 t2 ->
      subtype' (ty_list t1) (ty_list t2)

  | sub_listcons' : forall t1 t2,
      subtype' t1 t2 ->
      subtype' (ty_listcons t1) (ty_listcons t2)

  | sub_listmixed' : forall t1 t2,
      subtype' t1 t2 ->
      subtype' (ty_listcons t1) (ty_list t2)
  .

  Hint Constructors subtype' : typing.

  Lemma sub_trans : forall t1 t2 t3, subtype' t1 t2 -> subtype' t2 t3 -> subtype' t1 t3.
    intros t1 t2 t3 H; revert t3; induction H; intros; auto with typing.
    inversion H; subst; auto with typing.
    inversion H0; subst; auto with typing.
    inversion H0; subst; auto with typing.
    inversion H0; subst; auto with typing.
   Qed.

  Hint Resolve sub_trans : typing.

  Definition env_sub (e e':env) :=
    forall v t', e'#v = Some t' ->
       exists t, e#v = Some t /\ subtype' t t'.

  Lemma env_subtype_refl : forall e, env_sub e e.
    unfold env_sub; intros.
    exists t'; auto with typing.
  Qed.

  Lemma env_sub_trans : forall e1 e2 e3,
    env_sub e1 e2 ->
    env_sub e2 e3 ->
    env_sub e1 e3.

    intros; red; intros.
    elim (H0 _ _ H1).
    intros x [Hx1 Hx2].
    elim (H _ _ Hx1).
    intros y [Hy1 Hy2].
    exists y; split; auto.
    apply sub_trans with x; auto.
  Qed.

  Hint Resolve env_subtype_refl : typing.

  Inductive value_has_ty: value -> typ -> Prop :=
  | value_has_ty_nil_nil:
    value_has_ty (value_label (L 0)) ty_nil
  | value_has_ty_nil_list: forall t,
    value_has_ty (value_label (L 0)) (ty_list t)
  | value_has_ty_cons_listcons: forall v1 v2 t,
    value_has_ty v1 t ->
    value_has_ty v2 (ty_list t) ->
    value_has_ty (value_cons v1 v2) (ty_listcons t)
  | value_has_ty_cons_list: forall v1 v2 t,
    value_has_ty v1 t ->
    value_has_ty v2 (ty_list t) ->
    value_has_ty (value_cons v1 v2) (ty_list t).

  Hint Constructors value_has_ty: typing.

  Definition store_has_ty (s: store) (e: env) : Prop :=
    forall r t,
      e#r = Some t -> exists v, s#r = Some v /\ value_has_ty v t.

  Lemma subsumption:
    forall v t1, value_has_ty v t1 ->
      forall t2, subtype' t1 t2 -> value_has_ty v t2.

    intros v t1 H; induction H; intros.
    inversion H; auto with typing.
    inversion H; auto with typing.
    inversion H1; subst; clear H1; auto with typing.
    inversion H1; subst; clear H1; auto with typing.
  Qed.

  Lemma store_subsumption:
    forall r e, store_has_ty r e ->
      forall e', env_sub e e' -> store_has_ty r e'.

    intros; red; intros.
    red in H0.
    elim (H0 r0 t H1).
    intros x [Hx1 Hx2].
    red in H.
    elim (H r0 x Hx1).
    intros v [Hv1 Hv2].
    exists v; split; auto.
    apply subsumption with x; auto.
  Qed.

  Lemma var_set_type : forall s e r v t,
    store_has_ty s e ->
    value_has_ty v t ->
    store_has_ty (s#r <- v) (e#r <- t).

    red; intros.
    destruct (label_eq r r0).
    subst r0.
    rewrite get_set_same in H1.
    exists v.
    rewrite get_set_same.
    intuition.
    replace t0 with t; auto.
    congruence.
    rewrite get_set_other in H1; auto.
    rewrite get_set_other; auto.
  Qed.

  Lemma var_set_type_refine: forall s e v t r,
    store_has_ty s e ->
    s#r = Some v ->
    value_has_ty v t ->
    store_has_ty s (e#r <- t).

    red; intros.
    destruct (label_eq r r0).
    subst r0.
    rewrite get_set_same in H2.
    exists v; intuition.
    replace t0 with t; auto.
    congruence.
    rewrite get_set_other in H2; auto.
  Qed.

  Lemma var_unset_type : forall s e v x,
    store_has_ty s e ->
    store_has_ty (s#v <- x) (unset _ e v).

    red; intros.
    destruct (label_eq v r).
    subst v.
    rewrite unset_same in H0.
    discriminate.

    rewrite unset_other in H0; auto.
    red in H.
    destruct (H r t H0).
    exists x0; intuition.
    rewrite get_set_other; auto.
  Qed.

  Inductive check_instr': program_typing -> env -> instr -> env -> Prop :=

  | check_instr_seq': forall pt e e' e'' i1 i2,
    check_instr' pt e i1 e' ->
    check_instr' pt e' i2 e'' ->
    (*----------------------------------------------------*)
    check_instr' pt e (i1 ;; i2) e''

  | check_instr_branch_list': forall pt e e1 v l t,
    e#v = Some (ty_list t) ->
    pt#l = Some e1 ->
    env_sub (e#v <- ty_nil) e1 ->
    (*----------------------------------------------------*)
    check_instr' pt e (instr_branch_if_nil v l) (e#v <- (ty_listcons t))

  | check_instr_branch_listcons': forall pt e l v t,
    e#v = Some (ty_listcons t) ->
    (*----------------------------------------------------*)
    check_instr' pt e (instr_branch_if_nil v l) e

  | check_instr_branch_nil': forall pt e v l e1 e2,
    e#v = Some ty_nil ->
    pt#l = Some e1 ->
    env_sub e e1 ->
    (*----------------------------------------------------*)
    check_instr' pt e (instr_branch_if_nil v l) e2

  | check_instr_fetch_0': forall pt e v1 v2 t,
    e#v1 = Some (ty_listcons t) ->
    (*----------------------------------------------------*)
    check_instr' pt e (instr_fetch_field v1 0 v2) (e#v2 <- t)

  | check_instr_fetch_1': forall pt e v1 v2 t,
    e#v1 = Some (ty_listcons t) ->
    (*----------------------------------------------------*)
    check_instr' pt e (instr_fetch_field v1 1 v2) (e#v2 <- (ty_list t))

  | check_instr_cons': forall pt e v1 v2 v3 t1 t2 t3,
    e#v1 = Some t1 ->
    e#v2 = Some t2 ->
    subtype' t1 t3 ->
    subtype' t2 (ty_list t3) ->
    (*----------------------------------------------------*)
    check_instr' pt e (instr_cons v1 v2 v3) (e#v3 <- (ty_listcons t3)).

  Inductive check_block': program_typing -> env -> instr -> Prop :=

  | check_block_halt': forall pt e,
    (*----------------------------------------------------*)
    check_block' pt e instr_halt

  | check_block_seq': forall pt e i1 i2 e',
    check_instr' pt e i1 e' ->
    check_block' pt e' i2 ->
    (*----------------------------------------------------*)
    check_block' pt e (i1 ;; i2)

  | check_block_jump': forall pt e v l e1,
    pt#l = Some e1 ->
    env_sub (unset _ e v) e1 ->
    (*----------------------------------------------------*)
    check_block' pt e (instr_getlabel l v ;; instr_jump v).

  Hint Constructors check_instr' : typing.
  Hint Constructors check_block' : typing.

  Lemma set_sub_env_sub : forall e1 e2 v t1 t2,
    env_sub e1 e2 ->
    subtype' t1 t2 ->
    env_sub (e1#v <- t1) (e2#v <- t2).

    intros; red; intros.
    elim (var_eq v v0); intro Hv.
    subst v0.
    rewrite get_set_same in H1.
    replace t' with t2 in *; [ | congruence ].
    exists t1; split; auto.
    rewrite get_set_same; auto.

    rewrite get_set_other in H1; auto.
    red in H.
    destruct (H v0 t' H1); intuition.
    exists x.
    rewrite get_set_other; auto.
  Qed.

  Lemma set_same_sub : forall e1 e2 v t,
    env_sub e1 e2 ->
    env_sub (e1#v <- t) (e2#v <- t).

    intros; apply set_sub_env_sub; auto with typing.
  Qed.  

  Lemma check_instr_sub : forall pt e1 e1' i,
    check_instr' pt e1 i e1' ->
    forall e2, env_sub e2 e1 ->
      exists e2', check_instr' pt e2 i e2' /\ env_sub e2' e1'.

    intros pt e1 e1' i H; induction H; intros; auto with typing.

 (* check_instr_seq case *)
    elim (IHcheck_instr'1 e2 H1); intros x [Hx1 Hx2].
    elim (IHcheck_instr'2 x Hx2); intros y [Hy1 Hy2].
    exists y; split; auto.
    apply check_instr_seq' with x; auto.

 (* check_instr_branch_list case *)
    elim (H2 _ _ H).
    intros s [Hs1 Hs2].
    inversion Hs2; subst; clear Hs2; auto.

    exists (e2#v <- (ty_listcons t)); split; auto with typing.
    apply check_instr_branch_list' with e1; auto.
    apply env_sub_trans with (e#v<-ty_nil); auto.
    apply set_same_sub; auto.
    apply set_same_sub; auto.

    exists (e2#v <- (ty_listcons t)); split; auto with typing.
    apply check_instr_branch_nil' with e1; auto.
    apply env_sub_trans with (e#v<-ty_nil); auto.
    red; intros.
    elim (var_eq v v0); intros Hv.
    subst v0.
    exists ty_nil; split; auto.
    red in H2.
    rewrite get_set_same in H3.
    replace t' with ty_nil; [ | congruence ]; auto with typing.
    elim (H2 v0 t').
    intros z [Hz1 Hz2].
    exists z; auto.
    rewrite get_set_other in H3; auto.
    apply set_same_sub; auto.

    exists (e2#v <- (ty_listcons t1)); split; auto with typing.
    apply check_instr_branch_list' with e1; auto.
    apply env_sub_trans with (e#v<-ty_nil); auto.
    apply set_same_sub; auto.
    apply set_sub_env_sub; auto with typing.

    exists e2; split; auto.
    apply check_instr_branch_listcons' with t1; auto.
    red; intros.
    elim (var_eq v v0); intros Hv.
    subst v0.
    exists (ty_listcons t1); split; auto.
    rewrite get_set_same in H3.
    replace t' with (ty_listcons t); [ | congruence ]; auto with typing.
    rewrite get_set_other in H3; auto.

 (* check_instr_branch_listcons case *)
    elim (H0 _ _ H).
    intros s [Hs1 Hs2].
    inversion Hs2; subst; clear Hs2; auto.

    exists e2; split; auto.
    apply check_instr_branch_listcons' with t; auto.

    exists e2; split; auto.
    apply check_instr_branch_listcons' with t1; auto.

 (* check_instr_branch_nil case *)
    elim (H2 _ _ H).
    intros s [Hs1 Hs2].
    inversion Hs2; subst; clear Hs2; auto.
    exists e2; split; auto.
    apply check_instr_branch_nil' with e1; auto.
    apply env_sub_trans with e; auto.
    apply env_subtype_refl.

 (* check_instr_fetch_field_0 *)
    elim (H0 _ _ H).
    intros s [Hs1 Hs2].
    inversion Hs2; subst; clear Hs2; auto.
    exists (e2#v2 <- t); split; auto.
    apply check_instr_fetch_0'; auto.
    apply set_same_sub; auto.

    exists (e2#v2 <- t1); split; auto.
    apply check_instr_fetch_0'; auto.
    apply set_sub_env_sub; auto.

 (* check_instr_fetch_field_1 *)
    elim (H0 _ _ H).
    intros s [Hs1 Hs2].
    inversion Hs2; subst; clear Hs2; auto.
    exists (e2#v2 <- (ty_list t)); split; auto.
    apply check_instr_fetch_1'; auto.
    apply set_same_sub; auto.

    exists (e2#v2 <- (ty_list t1)); split; auto.
    apply check_instr_fetch_1'; auto.
    apply set_sub_env_sub; auto with typing.

 (* check_instr_cons *)
    elim (H3 _ _ H).
    intros s [Hs1 Hs2].
    elim (H3 _ _ H0).
    intros p [Hp1 Hp2].
    exists (e2#v3 <- (ty_listcons t3)); split; auto.
    eapply check_instr_cons' with s p; eauto with typing.
    apply set_sub_env_sub; auto with typing.
  Qed.

  Lemma check_block_sub : forall (p:program) pt e1 i,
    check_block' pt e1 i ->
    forall e2, env_sub e2 e1 ->
      check_block' pt e2 i.

    intros p pt e1 i H; induction H; intros; auto with typing.
    destruct (check_instr_sub pt e e' _ H e2); auto.
    intuition.
    apply check_block_seq' with x; auto.
    apply check_block_jump' with e1; auto.
    apply env_sub_trans with (unset _ e v); auto.
    red; intros.
    destruct (var_eq v v0).
    subst v0.
    rewrite unset_same in H2.
    discriminate.
    rewrite unset_other in H2; auto.
    rewrite unset_other; auto.
  Qed.   

  Definition check_instr := fun _:program => check_instr'.
  Definition check_block := fun _:program => check_block'.

  Definition check_blocks (p:program) (pt:program_typing) :=
   forall l i,
     p#l = Some i ->
     exists e,
       pt#l = Some e /\
       check_block p pt e i.

  Definition typing_dom_match (p: program) (pt: program_typing) : Prop :=
    forall l e, pt#l = Some e -> exists i, p#l = Some i.

  Definition check_program (p:program) (pt:program_typing) :=
    check_blocks p pt /\ typing_dom_match p pt /\ pt#entry_label = Some (empty _).

  Lemma preservation: forall p s1 i1 s1' i1',
    step p s1 i1 s1' i1' ->
    forall pt e1
      (CHKPROG: check_program p pt)
      (STORETY: store_has_ty s1 e1)
      (CHKBLOCK: check_block' pt e1 i1),
      exists s2, exists i2, exists n, n <= 1 /\
        stepN p n s1' i1' s2 i2 /\
        exists e2, store_has_ty s2 e2 /\ check_block' pt e2 i2.

    intros p s1 i1 s2 i2 H; induction H; intros.
    
  (* step_seq case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H3; subst; clear H3.
    exists r.
    exists (i1;;i2;;i3).
    exists 0.
    split; auto.
    split; [ apply stepN_O | ].
    exists e1; eauto with typing.

  (* step_fetch_field_0 case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H4; subst; clear H4.
    exists (r#v2 <- a0); exists i; exists 0;
      split; [ auto | ]; split; [ apply stepN_O | ].
    exists (e1#v2 <- t); split; auto.
    apply var_set_type; auto.
    elim STORETY with v1 (ty_listcons t); auto.
    intros x [Hx1 Hx2].
    replace x with (value_cons a0 a1) in *; auto.
    inversion Hx2; auto.
    congruence.

  (* step_fetch_field_1 case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H4; subst; clear H4.
    exists (r#v2 <- a1); exists i; exists 0;
      split; [ auto | ]; split; [ apply stepN_O | ].
    exists (e1#v2 <- (ty_list t)); split; auto.
    apply var_set_type; auto.
    elim STORETY with v1 (ty_listcons t); auto.
    intros x [Hx1 Hx2].
    replace x with (value_cons a0 a1) in *; auto.
    inversion Hx2; auto.
    congruence.

  (* step_cons case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H5; subst; clear H5.
    exists (r#v3 <- (value_cons a1 a2)); exists i; exists 0;
      split; [ auto | ]; split; [ apply stepN_O | ].
    exists (e1#v3 <- (ty_listcons t3)); split; auto.
    elim STORETY with v1 t1; auto.
    intros x [Hx1 Hx2].
    replace x with a1 in *; [ | congruence ].
    elim STORETY with v2 t2; auto.
    intros y [Hy1 Hy2].
    replace y with a2 in *; [ | congruence ].
    apply var_set_type; auto.
    apply value_has_ty_cons_listcons.
    apply subsumption with t1; auto with typing.
    apply subsumption with t2; auto.

  (* step_branch_not_taken case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H4; subst; clear H4.
    exists r; exists i; exists 0;
      split; [ auto | ]; split; [ apply stepN_O | ].
    exists (e1#v <- (ty_listcons t)); split; auto.
    apply var_set_type_refine with (value_cons a1 a2); auto with typing.
    elim STORETY with v (ty_list t); auto.
    intros x [Hx1 Hx2].
    replace x with (value_cons a1 a2) in *; [ | congruence ].
    inversion Hx2; subst; clear Hx2; auto.
    apply value_has_ty_cons_listcons; auto.

    exists r; exists i; exists 0;
      split; auto; split; [ apply stepN_O | ].
    exists e'; split; auto.

    elim STORETY with v ty_nil; auto.
    intros x [Hx1 Hx2]; inversion Hx2.
    replace x with (value_cons a1 a2) in *; [ | congruence ].
    discriminate.

  (* step_branch_taken case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H5; subst; clear H5.
    inversion CHKPROG; subst; clear CHKPROG.
    exists r; exists i'; exists 0; split; auto; split; [ apply stepN_O | ].
    exists e0; split; auto.
    apply store_subsumption with (e1#v <- ty_nil).
    apply var_set_type_refine with (value_label (L 0)); auto with typing.
    auto.
    destruct (H1 l i'); intuition.
    red in H9.
    replace e0 with x; [ auto | congruence ].

    elim STORETY with v (ty_listcons t); auto.
    intros x [Hx1 Hx2].
    replace x with (value_label (L 0)) in *; [ | congruence ].
    inversion Hx2.

    inversion CHKPROG; subst; clear CHKPROG.
    exists r; exists i'; exists 0; split; auto; split; [ apply stepN_O | ].
    exists e1; split; auto.
    apply check_block_sub with e0; auto.
    destruct (H1 l i'); intuition.
    red in H9.
    replace e0 with x; [ auto | congruence ].

  (* step_jump case -- can't happen in this typechecker *)
    inversion CHKBLOCK.

  (* getlabel then step_jump case *)
    inversion CHKBLOCK; subst; clear CHKBLOCK.
    inversion H3.
    inversion CHKPROG.
    exists (r#v <- (value_label l)).
    destruct H0.
    destruct (H0 l e0); auto.
    exists x.
    exists 1.
    split; auto.
    split.
    eapply stepN_S.
    eapply step_jump.
    rewrite get_set_same.
    reflexivity.
    apply H2.
    apply stepN_O.
    exists e0; split.
    apply store_subsumption with (unset typ e1 v).
    apply var_unset_type; auto.
    auto.
    destruct (H l x); intuition.
    red in H7.
    replace e0 with x0; [ auto | congruence ].
  Qed.

  Lemma progress: forall pt e i,
    check_block' pt e i ->
    forall p s,
      check_program p pt ->
      store_has_ty s e ->
      step_or_halt p s i.

    intros pt e i H; induction H; intros.
    
  (* halt case *)
    apply step_or_halt_halt.

  (* straightline instr *) 
    inversion H; subst; clear H.

  (* check_instr_seq case *)
    apply step_or_halt_step with s (i0 ;; (i3 ;; i2)).
    apply step_seq.

  (* check_instr_branch_list *)
    elim (H2 v (ty_list t)); auto.
    intros x [Hx1 Hx2].
    inversion Hx2; subst; clear Hx2.
    inversion H1; subst; clear H1.
    destruct H6.
    elim (H1 l e1); auto.
    intros i' Hi'.
    apply step_or_halt_step with s i'.
    apply step_branch_taken; auto.
    apply step_or_halt_step with s i2.
    eapply step_branch_not_taken.
    apply Hx1.

  (* check_instr_branch_listcons *)
    elim (H2 v (ty_listcons t)); auto.
    intros x [Hx1 Hx2].
    inversion Hx2; subst; clear Hx2.
    apply step_or_halt_step with s i2.
    eapply step_branch_not_taken.
    apply Hx1.

  (* check_instr_branch_nil *)
    elim (H2 v (ty_nil)); auto.
    intros x [Hx1 Hx2].
    inversion Hx2; subst; clear Hx2.
    inversion H1; subst; clear H1.
    destruct H6.
    elim (H1 l e1); auto.
    intros i' Hi'.
    apply step_or_halt_step with s i'.
    apply step_branch_taken; auto.

  (* check_instr_fetch_0 *)
    elim (H2 v1 (ty_listcons t)); auto.
    intros x [Hx1 Hx2].
    inversion Hx2; subst; clear Hx2.
    apply step_or_halt_step with (s#v2 <- v0) i2.
    apply step_fetch_field_0 with v3; auto.

  (* check_instr_fetch_1 *)
    elim (H2 v1 (ty_listcons t)); auto.
    intros x [Hx1 Hx2].
    inversion Hx2; subst; clear Hx2.
    apply step_or_halt_step with (s#v2 <- v3) i2.
    apply step_fetch_field_1 with v0; auto.

  (* check_instr_cons *)
    elim (H2 v1 t1); auto; intros x [Hx1 Hx2].
    elim (H2 v2 t2); auto; intros y [Hy1 Hy2].
    apply step_or_halt_step with (s#v3 <- (value_cons x y)) i2.
    apply step_cons; auto.

 (* jump case *)
    inversion H1; subst; clear H1.
    destruct H4.
    elim (H1 l e1); auto; intros i' Hi'.
    eapply step_or_halt_step.
    eapply step_getlabel.
  Qed.

  Section safety_aux.
    Variables
      (p : program)
      (pt : program_typing)
      (Hblk : check_blocks p pt)
      (Hdom : typing_dom_match p pt)
      (Hstart : pt # (L 0) = Some (empty typ)).

    Lemma safety_aux : forall n s i s' i',
      stepN p n s i s' i' ->
      forall m,
        check_block' pt m i ->
        store_has_ty s m ->
        exists s'', exists i'', exists n',
          n' <= 1 /\ stepN p n' s' i' s'' i'' /\
          exists e : env,
            store_has_ty s'' e /\
            check_block' pt e i'' /\
            step_or_halt p s'' i''.
      
      intro n; pattern n; apply CompleteInduction.
      clear n; intros n Hind.
      intros.

      inversion H; subst; clear H; simpl in *.
      
   (* base case *)
      clear Hind.
      exists s'; exists i'; exists 0; split; auto; split; [ apply stepN_O | ].
      exists m; intuition.
      apply progress with pt m; intuition.
      red; intuition.
      
   (* inductive case *)
      destruct (preservation _ _ _ _ _ H2 pt m) as [s''' [i''' [n' [Hn' [Hstp [e' ?]]]]]]
        ; intuition.
      red; intuition.
      destruct n0.
      inversion H3; subst.
      exists s'''; exists i'''; exists n'.
      split; auto.
      split; auto.
      exists e'; intuition.
      apply progress with pt e'; auto.
      red; intuition.   
      assert (n' <= S n0); [ omega | ].
      generalize (step_triangle p n' (S n0) s'0 s''' s' i'0 i''' i' H Hstp H3).
      intro H6.
      assert (S n0 - n' < S (S n0)); [ omega | ].
      apply (Hind _ H7 _ _ _ _ H6 _ H5 H4).
    Qed.

  End safety_aux.

  Lemma check_program_safety: forall p pt,
    check_program p pt ->
    safe_prog p.

    unfold safe_prog, safe, start_state; intros.
    destruct H as [Hblk [Hdom Hstart]].
    destruct (stepstar_stepN p s i s' i' H1) as [n Hn].
    assert (H2 : check_block' pt (empty typ) i).
    destruct (Hblk entry_label i); intuition.
    replace (empty typ) with x; [ auto | ].
    change typ with ty.
    congruence.
    assert (H3 : store_has_ty s (empty typ)).
    red; intros; destruct r; simpl in *; discriminate.
    destruct (safety_aux p pt Hblk Hdom Hstart n s i s' i' Hn (empty typ) H2 H3)
      as [s'' [i'' [n' [Hn' [Hstp [e' ?]]]]]]; intuition.
    destruct n'; inversion Hstp; subst; auto.
    eapply step_or_halt_step.
    apply H7.
  Qed.

  Theorem safety : forall
     (p:program) (pt: program_typing),
     check_blocks p pt ->
     typing_dom_match p pt ->
     pt#entry_label = Some (empty typ) ->
     safe_prog p.

     intros; apply check_program_safety with pt.
     red; intuition.
    Qed.

  Definition subtype := subtype'.
  Definition sub_refl := sub_refl'.
  Definition sub_nil := sub_nil'.
  Definition sub_list := sub_list'.
  Definition sub_listmixed := sub_listmixed'.
  Definition sub_listcons := sub_listcons'.
  Definition check_instr_seq := fun _:program => check_instr_seq'.
  Definition check_instr_branch_list := fun _:program => check_instr_branch_list'.
  Definition check_instr_branch_listcons := fun _:program => check_instr_branch_listcons'.
  Definition check_instr_branch_nil := fun _:program => check_instr_branch_nil'.
  Definition check_instr_fetch_0 := fun _:program => check_instr_fetch_0'.
  Definition check_instr_fetch_1 := fun _:program => check_instr_fetch_1'.
  Definition check_instr_cons := fun _:program => check_instr_cons'.
  Definition check_block_halt := fun _:program => check_block_halt'.
  Definition check_block_seq := fun _:program => check_block_seq'.
  Definition check_block_jump := fun _:program => check_block_jump'.
End SimpleTCSyntactic.

Module SyntacticSoundness :=
  SimpleTCSafetyTheorem SimpleTCSyntactic.

Check SyntacticSoundness.typecheck_safeprog.
